package sos.net;
/**
 * <p>Title: </p>
 * <p>Description: Mail-Client</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: SOS GmbH</p>
 * @author <a href="mailto:ghassan.beydoun@sos-berlin.com">Ghassan Beydoun</a>
 * @author Andreas Pschel <andreas.pueschel@sos-berlin.com>
 * @resource sos.util.jar sos.settings.jar sos.connection.jar sos.mail.jar sos.hostjava.jar hostjava.dll (Windows) libhostjava.so (Linux) libhostjava.sl (HP-UX) mail.jar smtp.jar activation.jar
 * @version $Id: SOSMail.java 606 2005-02-07 12:58:55Z gb $
 * @todo dequeue Mail mit Scheduler testen
 * @todo jz: content type
 * @todo jz: support for replyTo
 * @todo jz: replace forward slashes to backslashes in attachment names (is substituted to _)
 */

import javax.activation.DataHandler;
import javax.activation.FileDataSource;
import javax.mail.internet.InternetAddress;
import javax.mail.internet.MimeBodyPart;
import javax.mail.internet.MimeMessage;
import javax.mail.internet.MimeMultipart;
import javax.mail.Session;
import javax.mail.Transport;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;

import java.net.Socket;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Properties;


import sos.settings.SOSSettings;

import sos.util.SOSClassUtil;
import sos.util.SOSDate;


public class SOSMail {

    protected String host                 = "localhost";

    protected String port                 = "25";

    protected String user                 = "";

    protected String password             = "";

    protected int timeout                 = 15;

    protected String subject              = "";

    protected String from                 = "";
    
   protected String fromName              = "";

    protected String replyTo              = "";
    
    protected String queueDir             = "";
    
    protected String body                 = "";

    protected String language                = "de";

   protected String dateFormat                 = "dd.mm.yyyy";
   
   protected String datetimeFormat             = "dd.mm.yyyy HH:mm";
   
   protected HashMap dateFormats               = new HashMap();
   
   protected HashMap datetimeFormats           = new HashMap();

   /** Attribut: charSet: iso-8859-1, us-ascii */
   protected String charset               = "iso-8859-1";

    /** Mime-Type der Nachricht: text/plain, text/html etc.*/
    protected String contentType             = "text/plain";

   /** Attribut: encoding: Codierung des Nachrichteninhalts (Quoted-Printable, Base64) */
   protected String encoding              = "Base64";

   /** Mime-Type der Nachricht: application/octet-stream, application/pdf etc. */
    protected String attachmentContentType      = "application/octet-stream";

   /** Attribut: encoding: Codierung des Nachrichtenanhangs (Quoted-Printable, Base64) */
   protected String attachmentEncoding       = "Base64";

    protected String filename;

    protected boolean autoDequeue            = false;

    /** recipient Liste */
    protected Hashtable toList               = new Hashtable();

    /** cc Liste */
    protected Hashtable ccList               = new Hashtable();

    /** bcc Liste */
    protected Hashtable bccList           = new Hashtable();

    /** attachment Liste */
    protected HashMap attachmentList         = new HashMap();
    
    /** Muster fr eMail-Texte */
    protected Properties templates              = new Properties();

    /** sos settings object */
    protected SOSSettings sosSettings        = null;

    /** Tabelle fr Einstellungen */
    protected String tableSettings           = "SETTINGS";

    /** Applikationsname fr eMail-Einstellungen */
    protected String applicationMail         = "email"; 

    /** Sektionsname fr eMail-Einstellungen */
    protected String sectionMail             = "mail_server";
    
   /** Applikationsname fr eMail-Templates in Settings */
   protected String applicationMailTemplates    = "email_templates";
    
   /** Sektionsname fr eMail-Templates in Settings */
   protected String sectionMailTemplates     = "mail_templates";
    

    /**
     * Konstruktor
     * @param host string Hostname oder IP-Adresse des Mail-Servers
     * @throws java.lang.Exception
     */
    public SOSMail(String host) throws Exception {

        if (host != null) {
            this.host = host;
        }
        this.init();
    }


    /**
     * Konstruktor
     * @param host String Hostname oder IP-Adresse des Mail-Servers
     * @param user String Name des SMTP-Benutzers
     * @param pass String Kennwort des SMTP-Benutzers
     * @throws java.lang.Exception
     */
    public SOSMail(String host, String user, String password)
        throws Exception {

        if (host != null) {
            this.host = host;
        }
        if (user != null) {
            this.user = user;
        }
        if (password != null) {
            this.password = password;
        }
        this.init();
    }

   /**
    * Konstruktor
     * @param sosSettings SOSSettings Einstellungen aus Profile Settings oder Connection Settings
    * @throws java.lang.Exception
    */
   public SOSMail(SOSSettings sosSettings)
      throws Exception {

      this.getSettings(sosSettings);
      this.init();
   }
   
   /**
    * Konstruktor
     * @param sosSettings SOSSettings Einstellungen aus Profile Settings oder Connection Settings
    * @param language Sprache fr Einstellungen
    * @throws java.lang.Exception
    */
   public SOSMail(SOSSettings sosSettings, String language)
      throws Exception {

      this.getSettings(sosSettings, language);
      this.init();
   }
   
   /**
    * Initialisierungen
    * @throws java.lang.Exception
    */
   public boolean init()
      throws Exception {

      this.dateFormats.put( "de", "dd.MM.yyyy");
      this.dateFormats.put( "en", "MM/dd/yyyy");

      this.datetimeFormats.put( "de", "dd.MM.yyyy HH:mm");
      this.datetimeFormats.put( "en", "MM/dd/yyyy HH:mm");

      if ( this.dateFormats.containsKey(this.getLanguage()) && this.datetimeFormats.containsKey(this.getLanguage()) ) {
        this.setDateFormat( this.dateFormats.get(this.getLanguage()).toString());   
        this.setDatetimeFormat( this.datetimeFormats.get(this.getLanguage()).toString()); 
      } else {
        this.setDateFormat( this.dateFormats.get("de").toString() ); 
        this.setDatetimeFormat( this.datetimeFormats.get("de").toString() );  
      }

      return true;
   }
   
   /**
    * Einstellungen fr eMail-Versand aus Settings
    * 
    * @param sosSettings SOSSettings Einstellungen aus Profile Settings oder Connection Settings
    * @param language Sprache fr Einstellungen
    * @throws java.lang.Exception
    */
   public boolean getSettings(SOSSettings sosSettings, String language) throws Exception {

      if ( language != null ) { this.setLanguage(language); }
      return this.getSettings(sosSettings);
   }


    /**
    * Einstellungen fr eMail-Versand aus Settings
    * 
     * @param sosSettings SOSSettings Einstellungen aus Profile Settings oder Connection Settings
     * @throws java.lang.Exception
     */
    public boolean getSettings(SOSSettings sosSettings) throws Exception {

        if (sosSettings == null)
            throw new Exception(
                SOSClassUtil.getMethodName() + ": missing settings object.");

        this.sosSettings = sosSettings;

        Properties entries = new Properties();
      entries = this.sosSettings.getSection(this.applicationMail, this.sectionMail);

        if (entries.size() == 0)
            throw new Exception(
                SOSClassUtil.getMethodName()
                    + ": missing settings entries in section \"" + this.sectionMail + "\".");

        if (entries.containsKey("host") ) {
         if ( entries.getProperty("host").length() > 0 )
            this.host = entries.getProperty("host");
        }

        if (entries.containsKey("port") ) {
         if ( entries.getProperty("port").length() > 0 )
               this.port = entries.getProperty("port");
        }

      if (entries.containsKey("smtp_user") ) {
         if ( entries.getProperty("smtp_user").length() > 0 )
               this.user = entries.getProperty("smtp_user");
      }

      if ( entries.containsKey("smtp_pass") ) {
         if ( entries.getProperty("smtp_pass").length() > 0 ) 
            this.password = entries.getProperty("smtp_pass");
      }

        if (entries.getProperty("smtp_timeout") != null) {
         if ( entries.getProperty("smtp_timeout").length() > 0 )
               this.timeout = Integer.parseInt(entries.getProperty("smtp_timeout"));
        }

      if (entries.containsKey("from")) {
         if ( entries.getProperty("from").length() > 0 ) 
               this.from = entries.getProperty("from");
      }

        if (this.from == null && entries.containsKey("mail_from")) {
         if ( entries.getProperty("mail_from").length() > 0 )
               this.from = entries.getProperty("mail_from");
        }

      if (entries.containsKey("subject")) {
         if ( entries.getProperty("subject").length() > 0 )
               this.subject = entries.getProperty("subject");
      }

        if (entries.containsKey("language")) {
         if ( entries.getProperty("language").length() > 0 )
               this.language = entries.getProperty("language");
        }

        if (entries.containsKey("content_type")) {
         if ( entries.getProperty("content_type").length() > 0 )
               this.contentType = entries.getProperty("content_type");
        }

        if (entries.containsKey("encoding") ) {
         if ( entries.getProperty("encoding").length() > 0 )
               this.encoding = entries.getProperty("encoding");
        }

        if (entries.containsKey("charset")) {
         if ( entries.getProperty("charset").length() > 0 )
               this.charset = entries.getProperty("charset");
        }

      if (entries.containsKey("reply_to")) {
         if ( entries.getProperty("reply_to").length() > 0 )
            this.replyTo = entries.getProperty("reply_to");
      }

      if ( entries.containsKey("mail_queue_directory") ) {
         if ( entries.getProperty("mail_queue_directory").length() > 0 )
            this.queueDir = entries.getProperty("mail_queue_directory");
      }
        if (entries.containsKey("mail_auto_dequeue") ) {
         if ( entries.getProperty("mail_auto_dequeue").length() > 0 )
               this.autoDequeue =
                  Boolean
                    .valueOf(entries.getProperty("mail_auto_dequeue"))
                    .booleanValue();
        }
        return true;
    }

   /**
    * Muster fr eMail-Texte
    * 
    * @param sosSettings SOSSettings Einstellungen aus Profile Settings oder Connection Settings
    * @throws java.lang.Exception
    */
   public boolean getTemplates(SOSSettings sosSettings, String language) throws Exception {
      
      if ( language != null ) this.setLanguage(language);
      return this.getTemplates(sosSettings);
   }

   /**
    * Muster fr eMail-Texte
    * 
    * @param sosSettings SOSSettings Einstellungen aus Profile Settings oder Connection Settings
    * @throws java.lang.Exception
    */
   public boolean getTemplates(SOSSettings sosSettings) throws Exception {

      if (sosSettings == null)
         throw new Exception(
            SOSClassUtil.getMethodName() + ": missing settings object.");

      this.sosSettings = sosSettings;
      this.templates   = this.sosSettings.getSection(this.applicationMailTemplates, 
                                     this.sectionMailTemplates + "_" + language);

      if (this.templates.size() == 0)
         throw new Exception(
            SOSClassUtil.getMethodName()
               + ": missing settings entries for application \"" + this.applicationMailTemplates + "\" in section \"" + this.sectionMailTemplates + "\".");
      return true;
   }

   /**
    *
    * @param template Name des Templates
    * @param replacements HashMap mit Name/Wert-Paaren, der Name wird im Template durch den Wert ersetzt
    * @param nl2br NewLines durch HTML Breaks ersetzen
    * @throws java.lang.Exception
    */
   public String substituteSubject(String template, HashMap replacements) 
      throws Exception {

      if (!this.templates.containsKey(template + "_subject")) {
         throw (new Exception("substituteSubject(): template does not exist: " + template + "_subject" ));
      }
      
      return substitute( this.templates.get(template + "_subject").toString(), replacements, false );
   }

   /**
    *
    * @param template Name des Templates
    * @param replacements HashMap mit Name/Wert-Paaren, der Name wird im Template durch den Wert ersetzt
    * @param nl2br NewLines durch HTML Breaks ersetzen
    * @throws java.lang.Exception
    */
   public String substituteBody(String template, HashMap replacements, boolean nl2br) 
      throws Exception {

      if (!this.templates.containsKey(template + "_body")) {
         throw (new Exception("substituteBody(): template does not exist: " + template + "_body" ));
      }
      
      return substitute( this.templates.get(template + "_body").toString(), replacements, nl2br );
   }

   /**
    *
    * @param template Name des Templates
    * @param replacements HashMap mit Name/Wert-Paaren, der Name wird im Template durch den Wert ersetzt
    * @param nl2br NewLines durch HTML Breaks ersetzen
    * @throws java.lang.Exception
    */
   public String substituteBody(String template, HashMap replacements ) 
      throws Exception {

      if (! this.templates.containsKey(template + "_body")) {
         throw (new Exception("substituteBody(): template does not exist: " + template + "_body" ));
      }
      
      return substitute( this.templates.get(template + "_body").toString(), replacements, false);
   }


   /**
    *
    * @param content String mit beliebigem Inhalt
    * @param replacements HashMap mit Name/Wert-Paaren, der Name wird im content durch den Wert ersetzt
    * @param nl2br NewLines durch HTML Breaks ersetzen
    * @throws java.lang.Exception
    */
   public String substitute(String content, HashMap replacements, boolean nl2br) 
         throws Exception {

      Object key     = null;
      Object value   = null;

      content = content.replaceAll( "&\\(date\\)",     SOSDate.getCurrentDateAsString(this.dateFormat) );
      content = content.replaceAll( "&\\(datetime\\)", SOSDate.getCurrentTimeAsString(this.datetimeFormat) );
      
      if (nl2br) {
         content = content.replaceAll( "\n", "<br/>");
      }

        if ( replacements != null ) {
         Iterator keys  = replacements.keySet().iterator();
         while( keys.hasNext() ) {
            key = keys.next();
            if ( key != null ) {
               value = replacements.get(key.toString());
               if (value != null )
                  content = content.replaceAll( "&\\(" + key.toString() + "\\)", value.toString() );
            }
         }
        }

        // remove all variables that were not substituted
      return content.replaceAll( "&\\(.*\\)", "");
   }


    /**
     *
     * @param recipient eMail-Adresse des Empfngers
     * @throws java.lang.Exception
     */
    public void addRecipient(String recipient) throws Exception {

        if (recipient == null)
            throw new Exception(
                SOSClassUtil.getMethodName() + ": recipient has no value.");

        this.toList.put(recipient.toLowerCase(), recipient);
    }


    /**
     *
     * @param recipient eMail-Adresse des cc Empfngers
     * @throws java.lang.Exception
     */
    public void addCC(String recipient) throws Exception {

        if (recipient == null)
            throw new Exception(
                SOSClassUtil.getMethodName()
                    + ": CC recipient has no value.");

        this.ccList.put(recipient.toLowerCase(), recipient);
    }


    /**
     *
     * @param recipient eMail-Adresse des bcc Empfngers
     * @throws java.lang.Exception
     */
    public void addBCC(String recipient) throws Exception {

        if (recipient == null)
            throw new Exception(
                SOSClassUtil.getMethodName()
                    + ": BCC recipient has no value.");

        this.bccList.put(recipient.toLowerCase(), recipient);
    }


    /**
     *
     * @param filename Datei fr Anhang
     * @throws java.lang.Exception
     */
    public void addAttachment(String filename) throws Exception {

        if (filename == null)
            throw new Exception(
                SOSClassUtil.getMethodName() + ": attachment has no value.");

        String[] attachment = new String[2];
        attachment[0] = filename;
        attachment[1] = this.attachmentContentType;

        this.attachmentList.put(filename, attachment);
    }


    /**
     *
     * @param filename Datei fr Anhang
     * @param contentType Content-Type dieses Anhangs
     * @throws java.lang.Exception
     */
    public void addAttachment(String filename, String contentType)
        throws Exception {

        if (filename == null)
            throw new Exception(
                SOSClassUtil.getMethodName() + ": attachment has no value.");

        if (contentType == null) {
            contentType = this.attachmentContentType;
        }

        String[] attachment = new String[2];
        attachment[0] = filename;
        attachment[1] = contentType;

        this.attachmentList.put(filename, attachment);
    }


   /**
    * Nachricht versenden
    *
    * @throws java.lang.Exception
    */
   public void send() throws Exception {

      sendJavaMail();
   }


   /**
    * Nachricht mit JavaMail-Funktionen versenden
    *
    * @throws java.lang.Exception
    */
   public void sendJavaMail() throws Exception {

      SOSMailAuthenticator authenticator  = null;
      Session session               = null;
      MimeMessage message                 = null;
      MimeBodyPart attachedpart           = null;
      DataHandler dh                      = null;

      try {

         if (this.host == null || this.host.length() == 0) {
            throw new Exception(
               SOSClassUtil.getMethodName() + ": host has no value.");
         }

         if (this.port == null || this.port.length() == 0) {
            throw new Exception(
               SOSClassUtil.getMethodName() + ": port has no value.");
         }

         if (this.toList.isEmpty()) {
            throw new Exception(
               SOSClassUtil.getMethodName() + ": no recipient specified.");
         }

         if (this.from == null || this.from.length() == 0) {
            throw new Exception(
               SOSClassUtil.getMethodName() + ": no sender specified.");
         }

         Properties props = System.getProperties();
         props.put( "mail.host", this.host);
         props.put( "mail.user", this.user);
         props.put( "mail.transport.protocol", "smtp");
         props.put( "mail.smtp.class", "com.sun.mail.SMTPTransport");

         authenticator = new SOSMailAuthenticator();

         session = Session.getDefaultInstance(props, authenticator);

         message = new MimeMessage(session);
 
         if ( this.fromName != null && this.fromName.length() > 0 ) {
            message.setFrom( new InternetAddress(this.from, this.fromName) );
         } else {
            message.setFrom( new InternetAddress(this.from) );
         }

         if (this.replyTo != null && this.replyTo.length() > 0) {
            InternetAddress fromAddrs[] = new InternetAddress[1];
            fromAddrs[0] = new InternetAddress(this.replyTo);
            message.setReplyTo( fromAddrs );
         }


         if (!this.toList.isEmpty()) {
            InternetAddress toAddrs[] = new InternetAddress[toList.size()];
            int i = 0;           
            for (Enumeration e = this.toList.elements(); e.hasMoreElements();) {
               toAddrs[i++] = new InternetAddress(e.nextElement().toString());
            }
            message.addRecipients(MimeMessage.RecipientType.TO, toAddrs );
         }

         if (!this.ccList.isEmpty()) {
            InternetAddress toAddrs[] = new InternetAddress[ccList.size()];
            int i = 0;           
            for (Enumeration e = this.ccList.elements(); e.hasMoreElements();) {
               toAddrs[i++] = new InternetAddress(e.nextElement().toString());
            }
            message.addRecipients(MimeMessage.RecipientType.CC, toAddrs );
         }

         if (!this.bccList.isEmpty()) {
            InternetAddress toAddrs[] = new InternetAddress[bccList.size()];
            int i = 0;           
            for (Enumeration e = this.bccList.elements(); e.hasMoreElements(); ) {
               toAddrs[i++] = new InternetAddress(e.nextElement().toString());
            }
            message.addRecipients(MimeMessage.RecipientType.BCC, toAddrs );
         }

         if (this.subject != null) {
            message.setSubject(this.subject);
         }




         // send the attachments

         if (!this.attachmentList.isEmpty()) {
         
// Multipart nur bei Attachments!

           // send the body
           MimeBodyPart bodypart               = null;
           MimeMultipart multipart             = null;
        
           bodypart  = new MimeBodyPart();
           bodypart.setContent( this.body, this.contentType);
        
           if (this.filename != null && this.filename.length() > 0) {
              bodypart.setFileName(filename);
           }
        
           if (this.encoding != null) {
              // bodypart.setEncoding(this.encoding);
           }

           multipart = new MimeMultipart();
           multipart.addBodyPart(bodypart);
         
            for (Iterator i = this.attachmentList.values().iterator(); i.hasNext();) {
               String[] attachment = new String[2];
               attachment = (String[]) i.next();

               attachedpart  = new MimeBodyPart();
               FileDataSource fds = new FileDataSource( attachment[0] );
               if (1==0 && !attachment[1].equalsIgnoreCase("application/octet-stream")) {
                  dh = new DataHandler(fds, attachment[1]);
               } else {
                  dh = new DataHandler(fds);
               }
               attachedpart.setDataHandler(dh);
               attachedpart.setFileName( new File(attachment[0]).getName() );
               multipart.addBodyPart(attachedpart);
               message.setContent(multipart);
            }
         }else{
             message.setContent( this.body, this.contentType);
         }

         Transport.send(message);

      } catch (Exception e) {
         throw new Exception(
            SOSClassUtil.getMethodName() + ": error occurred on send: " + e.toString() );
      } finally {
      }
   }


    /**
     * Nachricht mit nativen Mail-Funktionen versenden
     *
     * @throws java.lang.Exception
     */
    public String sendNative() throws Exception {

        Socket socket   = null;
      String boundary = "DataSeparatorString";
        StringBuffer sb = new StringBuffer("");

        try {

         if (this.host == null || this.host.length() == 0) {
            throw new Exception(
               SOSClassUtil.getMethodName() + ": host has no value.");
         }

         if (this.port == null || this.port.length() == 0) {
            throw new Exception(
               SOSClassUtil.getMethodName() + ": port has no value.");
         }

         if (this.toList.isEmpty()) {
            throw new Exception(
               SOSClassUtil.getMethodName() + ": no recipient specified.");
         }

         if (this.from == null || this.from.length() == 0) {
            throw new Exception(
               SOSClassUtil.getMethodName() + ": no sender specified.");
         }

         socket = new Socket(this.host, Integer.parseInt(this.port));
         BufferedReader in = new BufferedReader
           (new InputStreamReader(socket.getInputStream(), "8859_1"));
         BufferedWriter out = new BufferedWriter
           (new OutputStreamWriter(socket.getOutputStream(), "8859_1"));

         sendLine(in, out, "HELO " + this.host);

         if ( this.fromName != null && this.fromName.length() > 0 
           && this.from != null && this.from.length() > 0 ) {
            sendLine(in, out, "MAIL FROM: " + this.fromName + getQuotedName(this.from));
         } else if ( this.from != null && this.from.length() > 0 ){
            sendLine(in, out, "MAIL FROM: " + getQuotedName(this.from));
         }

            if (!this.toList.isEmpty()) {
                for (Enumeration e = this.toList.elements(); e.hasMoreElements();) {
               sendLine(in, out, "RCPT TO: " + getQuotedName(e.nextElement().toString()) );
                }
            }

         sendLine(in, out, "DATA");
         sendLine(out, "MIME-Version: 1.0");

         if ( this.fromName != null && this.fromName.length() > 0 
           && this.from != null && this.from.length() > 0 ) {
            sendLine(out, "From: " + this.fromName + getQuotedName(this.from));
         } else if ( this.from != null && this.from.length() > 0 ){
            sendLine(out, "From: " + getQuotedName(this.from));
         }

         if (this.replyTo != null && this.replyTo.length() > 0) {
            sendLine(out, "Reply-To: " + getQuotedName(this.replyTo));
         }


         if (!this.toList.isEmpty()) {
            sb = new StringBuffer("");
            for (Enumeration e = this.toList.elements(); e.hasMoreElements();) {
               sb.append(getQuotedName(e.nextElement().toString()));
               if (e.hasMoreElements()) {
                  sb.append(",");
               }
            }
            sendLine(out, "To: " + sb);
         }

            if (!this.ccList.isEmpty()) {
            sb = new StringBuffer("");
                for (Enumeration e = this.ccList.elements(); e.hasMoreElements();) {
               sb.append(getQuotedName(e.nextElement().toString()));
                    if (e.hasMoreElements()) {
                        sb.append(",");
                    }
                }
            sendLine(out, "Cc: " + sb);
            }

            if (!this.bccList.isEmpty()) {
            sb = new StringBuffer("");
                for (Enumeration e = this.bccList.elements(); e.hasMoreElements(); ) {
                    sb.append(getQuotedName(e.nextElement().toString()));
                    if (e.hasMoreElements()) {
                  sb.append(",");
                    }
                }
            sendLine(out, "Bcc: " + sb);
            }

         if (this.subject != null) {
               sendLine(out, "Subject: " + this.subject);
         }

         sendLine(out, "Content-Type: multipart/mixed; boundary=\"" + boundary + "\"");
         sendLine(out, "\r\n--" + boundary);

         // send the body

            if (this.contentType != null && this.contentType.length() > 0) {
            sendLine(out, "Content-Type: text/html; charset=\"" + this.charset + "\"");
            }

         if (this.encoding != null) {
            sendLine(out, "Content-Transfer-Encoding: " + this.encoding);
         }

         sendLine(out, "\r\n" + this.body + "\r\n\r\n");

         // send the attachments

            if (!this.attachmentList.isEmpty()) {
                for (Iterator i = this.attachmentList.values().iterator();
                    i.hasNext();
                    ) {
                    String[] attachment = new String[2];
                    attachment = (String[]) i.next();

               sendLine(out, "\r\n--" +  boundary );
               sendLine(out, "Content-Type: " + attachment[1] + "; name=\"" + new File(attachment[0]).getName() + "\"");
               sendLine(out, "Content-Disposition: attachment; filename=\"" + attachment[0] + "\"");
               sendLine(out, "Content-Transfer-Encoding: " + this.attachmentEncoding + "\r\n");
               SOSMimeBase64.encode(attachment[0], out);
                }
            }

         sendLine(out, "\r\n\r\n--" + boundary + "--\r\n");
         sendLine(in, out, ".");
         sendLine(in, out, "QUIT");

        } catch (Exception e) {
         throw new Exception(
            SOSClassUtil.getMethodName() + ": error occurred on send: " + e.toString() );
        } finally {
         if ( socket != null ) { socket.close(); }
        }

      return sb.toString();
    }


   /**
    * Nachricht versenden mit hostWare
    *
    * @throws java.lang.Exception
    */
   public String sendHostware() throws Exception {

      sos.hostware.File file = new sos.hostware.File();
      StringBuffer sb;

      try {
         sb = new StringBuffer("-out mail ");

         if (this.host != null && this.port != null) {
            sb.append("-smtp=" + this.host + ":" + this.port + " ");
         } else if (this.host != null) {
            sb.append("-smtp=" + this.host + " ");
         }

         if ( this.fromName != null && this.fromName.length() > 0 
           && this.from != null && this.from.length() > 0 ) {
            sb.append("-from='" + this.fromName + "<" + this.from + ">' ");
         } else if ( this.from != null && this.from.length() > 0 ){
            sb.append("-from='" + this.from + "' ");
         }

         if (!this.toList.isEmpty()) {
            sb.append("-to='");
            for (Enumeration e = this.toList.elements();
               e.hasMoreElements();
               ) {
               sb.append(e.nextElement());
               if (e.hasMoreElements()) {
                  sb.append(",");
               }
            }
            sb.append("' ");
         }

         if (!this.ccList.isEmpty()) {
            sb.append("-cc='");
            for (Enumeration e = this.ccList.elements();
               e.hasMoreElements();
               ) {
               sb.append(e.nextElement());
               if (e.hasMoreElements()) {
                  sb.append(",");
               }
            }
            sb.append("' ");
         }

         if (!this.bccList.isEmpty()) {
            sb.append("-bcc='");
            for (Enumeration e = this.bccList.elements();
               e.hasMoreElements();
               ) {
               sb.append(e.nextElement());
               if (e.hasMoreElements()) {
                  sb.append(",");
               }
            }
            sb.append("' ");
         }

         if (this.replyTo != null) {
            sb.append("-reply-to='" + this.replyTo + "' ");
         }

         if (this.subject != null) {
            sb.append("-subject='" + this.subject + "' ");
         }

         if (this.filename != null) {
            sb.append("-filename='" + this.filename + "' ");
         }

         if (this.encoding != null) {
            sb.append("-encoding='" + this.encoding + "' ");
         }

         // not supported
         if (this.contentType != null && this.attachmentList.isEmpty()) {
            sb.append("-content-type='" + this.contentType + "' ");
         }

         if (this.autoDequeue ) {
            // automatic dequeueing not supported
            // sb.append("-dequeue-mail ");
         }

         if (this.queueDir != null) {
            sb.append("-queue-dir='" + this.queueDir + "' ");
         }

         if (!this.attachmentList.isEmpty()) {
            for (Iterator i = this.attachmentList.values().iterator();
               i.hasNext();
               ) {
               String[] attachment = new String[2];
               attachment = (String[]) i.next();
               sb.append(
                  "-attach='(-content-type="
                     + attachment[1]
                     + " "
                     + attachment[0]
                     + ")' ");
            }
         }

         file.open(sb.toString());
         file.put_line(this.body);
         file.close();

      } catch (Exception e) {
         throw new Exception(
            SOSClassUtil.getMethodName() + ": error occurred on send: " + e.toString() );
      } finally {
         if (file != null)
            file.destruct();
      }

      return sb.toString();
   }


   /**
    * setzt die eMail-Empfnger zurck
    */
   public void clearRecipients() {

      this.toList    = new Hashtable();
      this.ccList    = new Hashtable();
      this.bccList   = new Hashtable();
   }


   /**
    * setzt die eMail-Attachments zurck
    */
   public void clearAttachments() {

      this.attachmentList  = new HashMap();
   }


   /**
    * sendet eine Nachricht in den Ausgabestrom
    * @param out Ausgabestrom
    * @param s   beliebige Zeichenfolge
    */
   public void sendLine(BufferedReader in, BufferedWriter out, String s)
         throws Exception {

      try {
            out.write(s + "\r\n");
            out.flush();
            s = in.readLine();
         }
      catch (Exception e) {
         throw new Exception( SOSClassUtil.getMethodName() + 
                                 ": error occurred on sendLine: " + e.toString() );
         }
   }


   /**
    * sendet eine Nachricht in den Ausgabestrom
    * @param out Ausgabestrom
    * @param s   beliebige Zeichenfolge
    */
   public void sendLine(BufferedWriter out, String s)
          throws Exception {

      try {
            out.write(s + "\r\n");
            out.flush();
         }
      catch (Exception e) {
         throw new Exception( SOSClassUtil.getMethodName() + 
                         ": error occurred on sendLine: " + e.toString() );
         }
   }

   /**
    * liefert einen mit &lt; und &gt; quotierten String
    * @param host Hostname bzw. IP-Adresse des Mail-Servers
    */
   public String getQuotedName(String name) {

       if ( name.indexOf('<') > -1 && name.indexOf('>') > -1 ) {
         return name;
       } else {
         return '<' + name + '>';
       }
   }

    /**
    * setzt die IP-Adresse bzw. den Hostnamen des Mail-Servers
     * @param host Hostname bzw. IP-Adresse des Mail-Servers
     */
    public void setHost(String host) {
        this.host = host;
    }

   /**
    * liefert den Hostnamen des Mail-Servers
    */
   public String getHost() {
      return this.host;
   }

    /**
    * setzt den Port fr den Verbindungsaufbau zum Mail-Server
     * @param port Port-Nummer des Mail-Servers
     */
    public void setPort(String port) {
        this.port = port;
    }

   /**
    * liefert den Port fr den Verbindungsaufbau zum Mail-Server
    */
   public String getPort() {
      return this.port;
   }

    /**
    * setzt den Benutzernamen fr den Verbindungsaufbau zum Mail-Server bei Versand mit SMTP
     * @param user Name des SMTP-Users fr eMail-Versand
     */
    public void setUser(String user) {
        this.user = user;
    }

   /**
    * liefert den Benutzernamen fr den Verbindungsaufbau zum Mail-Server bei Versand mit SMTP
    */
   public String getUser() {
      return this.user;
   }

    /**
    * setzt das Kennwort fr den den Verbindungsaufbau zum Mail-Server bei Versand mit SMTP
     * @param password Kennwort des SMTP-Users fr eMail-Versand
     */
    public void setPassword(String password) {
        this.password = password;
    }

    /**
    * setzt den Timeout fr den Verbindungsaufbau zum Mail-Server
     * @param timeout fr Verbindungsaufbau zum Mail-Server (inaktiv)
     */
    public void setTimeout(int timeout) {
        this.timeout = timeout;
    }

   /**
    * liefert den Timeout fr den Verbindungsaufbau zum Mail-Server
    */
   public int getTimeout() {
      return this.timeout;
   }

    /**
    * liefert die Sprache fr eMail-Mustertexte und Datums-/Zeitformate
     */
    public String getLanguage() {
        return this.language;
    }

   /**
    * setzt die Sprache fr eMail-Mustertexte und Datums-/Zeitformate
    * @param language Sprache fr eMail-Muster (en, de, default:de)
    */
   public void setLanguage(String language) throws Exception {
      this.language = language;
      this.init();
   }

   /**
    * liefert das Datumsformat fr Datumsersetzungen im Text der eMail
    */
   public String getDateFormat() {
      return this.dateFormat;
   }

   /**
    * setzt das Datumsformat fr Datumsersetzungen im Text der eMail
    * @param dateFormat Datumsformat, z.B. dd.MM.yyyy
    */
   public void setDateFormat(String dateFormat) {
      this.dateFormat = dateFormat;
   }

   /**
    * liefert das Datums- und Zeitformat fr Datumsersetzungen im Text der eMail
    */
   public String getDatetimeFormat() {
      return this.datetimeFormat;
   }

   /**
    * setzt das Datums- und Zeitformat fr Datumsersetzungen im Text der eMail
    * @param datetimeFormat Datumsformat, z.B. dd.MM.yyyy HH:mm:ss
    */
   public void setDatetimeFormat(String datetimeFormat) {
      this.datetimeFormat = datetimeFormat;
   }

    /**
     * setzt das Encoding der eMail
     * @param encoding
     */
    public void setEncoding(String encoding) {
        this.encoding = encoding;
    }

   /**
    * liefert das Encoding der eMail
    */
   public String getEncoding() {
      return this.encoding;
   }

    /**
    * setzt das Character-Set der eMail
     * @param charset
     */
    public void setCharset(String charset) {
        this.charset = charset;
    }

   /**
    * liefert das Character-Set der eMail
    */
   public String getCharset() {
      return this.charset;
   }

    /**
     * setzt den Content-Type der eMail
     * @param contentType Content-Type der eMail
     */
    public void setContentType(String contentType) {
        this.contentType = contentType;
    }

   /**
    * liefert den Content-Type der eMail
    */
   public String getContentType() {
      return this.contentType;
   }

   /**
    * setzt den Content-Type fr Attachments
    * @param attachmentContentType Default fr Content-Type der Attachments
    */
   public void setAttachmentContentType(String attachmentContentType) {
      this.attachmentContentType = attachmentContentType;
   }

   /**
    * liefert den Content-Type fr Attachments
    */
   public String getAttachmentContentType() {
      return this.attachmentContentType;
   }

    /**
     * setzt den Namen des Verzeichnisses fr zwischengespeicherte eMails
     * @param queueDir String Verzeichnis zum Zwischenspeichern von eMails
     */
    public void setQueueDir(String queueDir) {
        this.queueDir = queueDir;
    }

   /**
    * liefert den Namen des Verzeichnisses fr zwischengespeicherte eMails
    */
   public String getQueueDir() {
      return this.queueDir;
   }

    /**
     * setzt den Zustand, dass automatischer Versand zwischengespeicherter eMails aktiviert ist
     * @param autoDequeue Flat zur automatischen Prfung und erneutem Versand zwischengespeicherter Mails
     */
    public void setAutoDequeue(boolean autoDequeue) {
        this.autoDequeue = autoDequeue;
    }

   /**
    * liefert den Zustand, ob automatischer Versand zwischengespeicherter eMails aktiviert ist
    */
   public boolean getAutoDequeue() {
      return this.autoDequeue;
   }

    /**
     * setzt den Inhalt des Betreffs
     * @param subject Betreff
     */
    public void setSubject(String subject) {
        this.subject = subject;
    }

   /**
    * liefert den Inhalt des Betreffs
    */
   public String getSubject() {
      return this.subject;
   }

    /**
     * setzt die eMail-Adresse des Absenders
     * @param from Absender der eMail
     */
    public void setFrom(String from) {
        this.from = from;
    }

   /**
    * liefert die eMail-Adresse des Absenders
    */
   public String getFrom() {
      return this.from;
   }

   /**
    * liefert den Namen des Absenders
    */
   public String getFromName() {
      return this.fromName;
   }

   /**
    * setzt den Namen des Absenders
    * @param fromName Absender der eMail
    */
   public void setFromName(String fromName) {
      this.fromName = fromName;
   }

    /**
     * setzt die eMail-Adresse fr Rckantworten
     * @param replyTo Empfnger fr Rckantworten
     */
    public void setReplyTo(String replyTo) {
        this.replyTo = replyTo;
    }

   /**
    * liefert die eMail-Adresse fr Rckantworten
    */
   public String getReplyTo() {
      return this.replyTo;
   }

    /**
     * setzt den Name einer Datei, die den eMail-Body beinhaltet
     * @param filename Name einer Datei, in der der eMail-Body als Anhang aufgenommen wird
     */
    public void setFilename(String filename) {
        this.filename = filename;
    }

    /**
      * setzt den Inhalt des eMail-Texts
      * @param body Inhalt der eMail
      */
    public void setBody(String body) {
        this.body = body;
    }

   /**
    * liefert den Inhalt des eMail-Texts 
    */
   public String getBody() {
      return this.body;
   }

   /**
    * setzt den Namen der Applikation fr eMail-Einstellungen in der Tabelle der Einstellungen
    * @param applicationMail Name der Applikation fr eMail-Einstellungen in der Einstellungstabelle
    */
   public void setApplicationMail(String applicationMail) {
      this.applicationMail = applicationMail;
   }

   /**
    * liefert den Namen der Applikation fr eMail-Einstellungen in der Tabelle der Einstellungen
    */
   public String getApplicationMail() {
      return this.applicationMail;
   }

   /**
    * setzt den Namen der Sektion fr eMail-Einstellungen in der Tabelle der Einstellungen
    * @param sectionMail Name der Sektion fr eMail-Einstellungen in der Einstellungstabelle
    */
   public void setSectionMail(String sectionMail) {
      this.sectionMail = sectionMail;
   }

   /**
    * liefert den Namen der Sektion fr eMail-Einstellungen in der Tabelle der Einstellungen
    */
   public String getSectionMail() {
      return this.sectionMail;
   }

   /**
    * setzt den Namen der Applikation fr Mustertexte in der Tabelle der Einstellungen
    * @param applicationMailTemplates Name der Applikation fr eMail-Muster in der Einstellungstabelle
    */
   public void setApplicationMailTemplates(String applicationMailTemplates) {
      this.applicationMailTemplates = applicationMailTemplates;
   }

   /**
    * liefert den Namen der Applikation fr Mustertexte in der Tabelle der Einstellungen
    */
   public String getApplicationMailTemplates() {
      return this.applicationMailTemplates;
   }

   /**
    * setzt den Namen der Sektion fr Mustertexte in der Tabelle der Einstellungen
    * @param sectionMailTemplates Name der Sektion fr eMail-Muster in der Einstellungstabelle
    */
   public void setSectionMailTemplates(String sectionMailTemplates) {
      this.sectionMailTemplates = sectionMailTemplates;
   }

   /**
    * liefert den Namen der Sektion fr Mustertexte in der Tabelle der Einstellungen
    */
   public String getSectionMailTemplates() {
      return this.sectionMailTemplates;
   }

   /**
    * liefert den Namen der Tabelle mit Einstellungen
    * @param tableSettings
    */
   public void setTableSettings(String tableSettings) {
      this.tableSettings = tableSettings;
   }

   /**
    * liefert den Namen der Tabelle mit Einstellungen
    */
   public String getTableSettings() {
      return this.tableSettings;
   }

}
