/*
 * ManagedReporter.java
 * Created on 26.10.2005
 * 
 */
package sos.scheduler.managed;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Properties;
import java.util.Set;
import java.util.Vector;

import sos.net.SOSMail;
import sos.settings.SOSSettings;
import sos.spooler.Task;
import sos.spooler.Variable_set;
import sos.util.SOSDate;
import sos.util.SOSLogger;

/**
 * Hilfsklasse fr Jobs, die einen Report verschicken wollen
 *
 * @author Andreas Liebert 
 */
public class ManagedReporter {
	
	// Einstellungen des Jobs, der den Reporter verwendet
	private SOSSettings jobSettings;
	
	// Logger des Jobs
	private SOSLogger logger;
	
	// Mail Objekt zum verschicken der Reports
	private SOSMail mail;
	
	// Task des aufrufenden Jobs
	private Task spooler_task;
	
	// aufrufender Job
    private JobSchedulerManagedJob job;
    
	// Luft der Job mit einem Order?
	private boolean orderJob;
	
	// Hashmap mit Ersetzugen fr body, subject...
	private HashMap replacements = new HashMap();
	
	// body Vorlage
	private String body = "";
	
	// subject Vorlage
	private String subject = "Report [taskid]";
	
	// Payload des Auftrags
	private Variable_set orderPayload;
	
	// Report file
	private File attach;
	
	// sollen Reports auch ohne Ergebnis verschickt werden?
	private boolean sendIfNoResult = true;
	
	// sollen Reports, die ein Ergebnis haben verschickt werden?
	private boolean sendIfResult = true;
	
	// hat der aktuelle report ein Ergebnis?
	private boolean hasResult = true;
	
	public ManagedReporter(JobSchedulerManagedJob job) throws Exception{
		this.setJobSettings(job.getJobSettings());
		this.setLogger(job.getLogger());
		this.job=job;
		spooler_task = job.spooler_task;
		String mailServer="";
		String logMailFrom = "";
		try{
			Properties spoolProp = getJobSettings().getSection("spooler");
			mailServer = spoolProp.getProperty("smtp");
			logMailFrom = spoolProp.getProperty("log_mail_from");
			//mailServer = spooler_log.mail().smtp();
			//getLogger().debug9("mail().smtp():"+mailServer);
			//logMailFrom = spooler_log.mail().from();
			//getLogger().debug9("mail().from():"+logMailFrom);
		} catch (Exception e){}
		try{
			mail = new SOSMail(mailServer);
			getLogger().debug9("Setting mail sender: "+logMailFrom);
			mail.setFrom(logMailFrom);
			mail.setCharset("ISO-8859-1");
		} catch(Exception e){
			throw new Exception("Failed to initialize SOSMail class."+e,e);
		}
	}
	
	public File getReportFile() throws Exception{
		orderJob = !(spooler_task.job().order_queue() == null);
		
		String fileName="report_[date]_[taskid].xml";
		orderPayload=job.getOrderPayload();
		
				
		try{
			
			if (getOrderPayload()!=null && getOrderPayload().var("scheduler_order_report_mailto")!=null &&
					getOrderPayload().var("scheduler_order_report_mailto").length()>0){
				String[] mails = getOrderPayload().var("scheduler_order_report_mailto").split(",");
				for (int i=0; i<mails.length; i++){
					mail.addRecipient(mails[i].trim());
				}			
				job.debugParamter(getOrderPayload(),"scheduler_order_report_mailto");
			}
			if (getOrderPayload()!=null && getOrderPayload().var("scheduler_order_report_mailcc")!=null &&
					getOrderPayload().var("scheduler_order_report_mailcc").length()>0){
				String[] mails = getOrderPayload().var("scheduler_order_report_mailcc").split(",");
				for (int i=0; i<mails.length; i++){
					mail.addCC(mails[i].trim());
				}
				job.debugParamter(getOrderPayload(),"scheduler_order_report_mailcc");
			}
			if (getOrderPayload()!=null && getOrderPayload().var("scheduler_order_report_mailbcc")!=null &&
					getOrderPayload().var("scheduler_order_report_mailbcc").length()>0){
				String[] mails = getOrderPayload().var("scheduler_order_report_mailbcc").split(",");
				for (int i=0; i<mails.length; i++){
					mail.addBCC(mails[i].trim());
				}			
				job.debugParamter(getOrderPayload(),"scheduler_order_report_mailbcc");
			}
			if (getOrderPayload()!=null && getOrderPayload().var("scheduler_order_report_subject")!=null &&
					getOrderPayload().var("scheduler_order_report_subject").length()>0){
				subject=(getOrderPayload().var("scheduler_order_report_subject"));
				job.debugParamter(getOrderPayload(),"scheduler_order_report_subject");
			}
			subject = replace(subject);
			if (getOrderPayload()!=null && getOrderPayload().var("scheduler_order_report_body")!=null &&
					getOrderPayload().var("scheduler_order_report_body").length()>0){
				body=(getOrderPayload().var("scheduler_order_report_body"));
				job.debugParamter(getOrderPayload(),"scheduler_order_report_body");
			}
			body=replace(body);
			if (getOrderPayload()!=null && getOrderPayload().var("scheduler_order_report_filename")!=null &&
					getOrderPayload().var("scheduler_order_report_filename").length()>0){
				fileName=getOrderPayload().var("scheduler_order_report_filename");
				job.debugParamter(getOrderPayload(),"scheduler_order_report_filename");			
			}	
			
			fileName=replace(fileName);
			getLogger().debug6("Output filename: "+fileName);
			if (getOrderPayload()!=null && getOrderPayload().var("scheduler_order_report_path")!=null && 
					getOrderPayload().var("scheduler_order_report_path").length()>0){
				String path = getOrderPayload().var("scheduler_order_report_path");
				job.debugParamter(getOrderPayload(),"scheduler_order_report_path");
				if (!path.endsWith("/") && !path.endsWith("\\") && path.length()>0) path = path +"/";
				path=replace(path);
				File reportPath = new File(path);
				if(!reportPath.exists()) reportPath.mkdirs();
				attach = new File(path+fileName);
			}else{
				String path = System.getProperty("java.io.tmpdir");
				if (!path.endsWith("/") && !path.endsWith("\\") && path.length()>0) path = path +"/";
				attach = new File(path+fileName);
				attach.deleteOnExit();
			}
			if (getOrderPayload()!=null && getOrderPayload().var("scheduler_order_report_send_if_no_result")!=null &&
					getOrderPayload().var("scheduler_order_report_send_if_no_result").length()>0){
				sendIfNoResult=(getOrderPayload().var("scheduler_order_report_send_if_no_result").equals("1") ||
						getOrderPayload().var("scheduler_order_report_send_if_no_result").equalsIgnoreCase("true"));
				job.debugParamter(getOrderPayload(),"scheduler_order_report_send_if_no_result");			
			}
			if (getOrderPayload()!=null && getOrderPayload().var("scheduler_order_report_send_if_result")!=null &&
					getOrderPayload().var("scheduler_order_report_send_if_result").length()>0){
				sendIfResult=(getOrderPayload().var("scheduler_order_report_send_if_result").equals("1") ||
						getOrderPayload().var("scheduler_order_report_send_if_result").equalsIgnoreCase("true"));
				job.debugParamter(getOrderPayload(),"scheduler_order_report_send_if_result");			
			}
			
			return attach;
		} catch(Exception e){
			mail.clearAttachments();
			mail.clearRecipients();
			clearReplacements();
			throw new Exception("error occurred preparing report file: "+e,e);
		}
	}
	
	public void report() throws Exception{
		try{
		    if(hasResult() && !sendIfResult){
		   	  getLogger().info("Query returned a result. No report will be sent because " +
		   	  		"scheduler_order_send_if_result=false");
		   	  return;
		    }
		    if(!hasResult() && !sendIfNoResult){
		    	getLogger().info("Query returned no result. No report will be sent because " +
	   	  		"scheduler_order_send_if_no_result=false");
		    	return;
		    }
		
			boolean asbody=false;
				
		if (getOrderPayload()!=null && getOrderPayload().var("scheduler_order_report_asbody")!=null){
			asbody= (
					getOrderPayload().var("scheduler_order_report_asbody").equals("1") ||
					getOrderPayload().var("scheduler_order_report_asbody").equalsIgnoreCase("true"));
			job.debugParamter(getOrderPayload(),"scheduler_order_report_asbody");
		}
		if(asbody){
			BufferedReader reader = new BufferedReader(new FileReader(attach));
			StringBuffer buffer = new StringBuffer();
			String line = reader.readLine();
			while (line != null) {
				buffer.append(line);
				buffer.append('\n');
				line = reader.readLine();
			}
			reader.close();
			body = buffer.toString();
			mail.setContentType("text/html");
		}
		if(body.trim().startsWith("<html")) mail.setContentType("text/html");
		mail.setBody(body);	
		mail.setSubject(subject);
		if (!asbody) mail.addAttachment(attach.getAbsolutePath());
		getLogger().info("Sending report email...");
		mail.send();
		} catch (Exception e){
			throw new Exception("Error occurred sending report: "+e,e);
		}finally{
			mail.clearAttachments();
			mail.clearRecipients();
			clearReplacements();
		}
	}
	
	private String replace(String source){
		String target;
		try{
			target = source.replaceAll("\\[date\\]",SOSDate.getCurrentDateAsString());
			target = target.replaceAll("\\[datetime\\]",SOSDate.getCurrentTimeAsString());
            target = source.replaceAll("\\[date_german\\]",SOSDate.getCurrentDateAsString("dd.MM.yyyy"));
            target = target.replaceAll("\\[datetime_german\\]",SOSDate.getCurrentTimeAsString("dd.MM.yyyy HH:mm:ss"));
			target = target.replaceAll("\\[subject\\]", this.getSubject());
            target = target.replaceAll("\\[nl\\]", "\n");
			if(isOrderJob())
			target = target.replaceAll("\\[orderid\\]", spooler_task.order().id());
			target = target.replaceAll("\\[jobname\\]", spooler_task.job().name());
			target = target.replaceAll("\\[taskid\\]", Integer.toString(spooler_task.id()));
			/* target = target.replaceAll("\\[sql\\]",sql);
			target = target.replaceAll("\\[xml\\]",xml); */
			Set keys = replacements.keySet();
			Iterator keysIt = keys.iterator();
			while(keysIt.hasNext()){
				String key = keysIt.next().toString();
				String repl = replacements.get(key).toString();
				target=target.replaceAll(key,repl);
			}
			
		} catch (Exception e){
			try{
				getLogger().warn("An error occurred replacing fields in String \""+source+"\"");
			} catch(Exception ex){}
			return source;
		}
		return target;
	}
	
	/**
	 * Fgt eine weitere Ersetzung hinzu
	 * @param regex Regulrer Ausdruck der ersetzt werden soll
	 * @param replacement String mit dem Ersetzungstext
	 */
	public void addReplacement(String regex, String replacement){		
		replacements.put(regex, replacement);		
	}
	
	
	/**
	 * lscht die Ersetzungen
	 */
	public void clearReplacements(){
		replacements.clear();
	}
	
	/**
	 * @return Returns the logger.
	 */
	public SOSLogger getLogger() {
		return logger;
	}
	/**
	 * @param logger The logger to set.
	 */
	public void setLogger(SOSLogger logger) {
		this.logger = logger;
	}
	/**
	 * @return Returns the jobSettings.
	 */
	public SOSSettings getJobSettings() {
		return jobSettings;
	}
	/**
	 * @param jobSettings The jobSettings to set.
	 */
	public void setJobSettings(SOSSettings jobSettings) {
		this.jobSettings = jobSettings;
	}
	/**
	 * @return Returns the orderJob.
	 */
	public boolean isOrderJob() {
		return orderJob;
	}
	/**
	 * @return Returns the body.
	 */
	public String getBody() {
		return body;
	}
	/**
	 * @param body The body to set.
	 */
	public void setBody(String body) {
		this.body = body;
	}
	/**
	 * @return Returns the subject.
	 */
	public String getSubject() {
		return subject;
	}
	/**
	 * @param subject The subject to set.
	 */
	public void setSubject(String subject) {
		this.subject = subject;
	}
	/**
	 * @return Returns the orderPayload.
	 */
	public Variable_set getOrderPayload() {
		return orderPayload;
	}
	/**
	 * @return Returns the hasResult.
	 */
	public boolean hasResult() {
		return hasResult;
	}
	/**
	 * @param hasResult The hasResult to set.
	 */
	public void setHasResult(boolean hasResult) {
		this.hasResult = hasResult;
	}
}
