package sos.util;

import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.Locale;

/**
 * Title: SOSDate
 * 
 * @author <a href="mailto:ghassan.beydoun@sos-berlin.com">Ghassan Beydoun </a>
 * @author Andreas P?schel
 * @author Titus Meyer
 * @version $Id: SOSDate.java 1432 2005-10-11 12:10:40Z re $
 */

public class SOSDate {

  /** Standard Datumsformat (ISO) */
  private static String   dateFormat     = new String("yyyy-MM-dd");

  /** Standard Zeitformat (ISO) */
  private static String   dateTimeFormat = new String("yyyy-MM-dd HH:mm:ss");
  
  /** Standard Ausgabe Zeitformat */
  private static String   outputDateTimeFormat = new String("MM/dd/yy HH:mm:ss");
  
  /** Constant of java.text.DateFormat for short style pattern */
  public static final int SHORT          = DateFormat.SHORT;

  /** Constant of java.text.DateFormat for medium style pattern */
  public static final int MEDIUM         = DateFormat.MEDIUM;

  /** Constant of java.text.DateFormat for long style pattern */
  public static final int LONG           = DateFormat.LONG;

  /** Constant of java.text.DateFormat for full style pattern */
  public static final int FULL           = DateFormat.FULL;

 	/** Datumsformat (default Jahr 2.stellig) */
  public static int dateStyle      			= DateFormat.SHORT;
  
  /** Zeitformat (default ohne Millisekunden) */
  public static int timeStyle      			= DateFormat.SHORT;
  
  /** Intenationalisierung */
  public static Locale locale      			= Locale.UK;
  
 

  /**
   * setzt das Datumsformat f?r alle Methoden dieser Klasse (default: ISO-Format
   * yyyy-mm-dd.
   * 
   * @param dateFormat Datumsformat abweichend von ISO
   */
  public void setDateFormat(String dateFormat) {

    SOSDate.dateFormat = dateFormat;
  }


  /**
   * liefert das Datumsformat (default: ISO-Format yyyy-mm-dd).
   */
  public static String getDateFormat() {

    return SOSDate.dateFormat;
  }


  /**
   * setzt das Zeitformat f?r alle Methoden dieser Klasse (default: ISO-Format
   * yyyy-mm-dd HH:mm:ss).
   * 
   * @param dateTimeFormat Zeitformat abweichend von ISO
   */
  public static void setDateTimeFormat(String dateTimeFormat) {

    SOSDate.dateTimeFormat = dateTimeFormat;
  }


  /**
   * liefert das Zeitformat (default: ISO-Format yyyy-mm-dd hh:mm:ss).
   */
  public static String getDateTimeFormat() {

    return SOSDate.dateTimeFormat;
  }


  /**
   * liefert das aktuelle Datum als Date zur?ck.
   * 
   * @exception Exception wird ausgel?st, falls ein Fehler vorliegt.
   */
  public static Date getCurrentDate() throws Exception {

    return SOSDate.getDate();
  }


  /**
   * liefert das aktuelle Datum als String im ISO-Format zur?ck.
   * 
   * @exception Exception wird ausgel?st, falls ein Fehler vorliegt.
   */
  public static String getCurrentDateAsString() throws Exception {

    SimpleDateFormat formatter = new SimpleDateFormat(SOSDate.dateFormat);
    Calendar now = Calendar.getInstance();
    return formatter.format(now.getTime());
  }


  /**
   * liefert das aktuelle Datum als String im angegebenen Format zur?ck.
   * 
   * @param dateFormat Datumsformat abweichend von ISO
   * @exception Exception wird ausgel?st, falls ein Fehler vorliegt.
   */
  public static String getCurrentDateAsString(String dateFormat)
      throws Exception {

    SimpleDateFormat formatter = new SimpleDateFormat(dateFormat);
    Calendar now = Calendar.getInstance();
    return formatter.format(now.getTime());
  }


  /**
   * liefert die aktuelle Uhrzeit als Date zur?ck.
   * 
   * @exception Exception wird ausgel?st, falls ein Fehler vorliegt.
   */
  public static Date getCurrentTime() throws Exception {

    return SOSDate.getTime();
  }


  /**
   * liefert die aktuelle Uhrzeit als String im default-Format (ISO) zur?ck.
   * 
   * @exception Exception wird ausgel?st, falls ein Fehler vorliegt.
   */
  public static String getCurrentTimeAsString() throws Exception {

    SimpleDateFormat formatter = new SimpleDateFormat(SOSDate.dateTimeFormat);
    Calendar now = Calendar.getInstance();
    return formatter.format(now.getTime());
  }


  /**
   * liefert die aktuelle Uhrzeit als String im angegebenen Format zur?ck.
   * 
   * @param dateTimeFormat Format-String, z.B. yyyy-mm-dd hh:mm:ss
   * @exception Exception wird ausgel?st, falls ein Fehler vorliegt.
   */
  public static String getCurrentTimeAsString(String dateTimeFormat)
      throws Exception {

    SimpleDateFormat formatter = new SimpleDateFormat(dateTimeFormat);
    Calendar now = Calendar.getInstance();
    return formatter.format(now.getTime());
  }


  /**
   * liefert das aktuelle Datum als Date zur?ck.
   * 
   * @exception Exception wird ausgel?st, falls ein Fehler vorliegt.
   */
  public static Date getDate() throws Exception {

    SimpleDateFormat formatter;

    try {
      formatter = new SimpleDateFormat(SOSDate.dateFormat);
    } catch (Exception e) {
      throw (new Exception("invalid date format string: " + e.toString()));
    }

    try {
      Calendar now = Calendar.getInstance();
      return now.getTime();
      // return formatter.parse(now.getTime().toString());
    } catch (Exception e) {
      throw (new Exception("illegal date value: " + e.toString()));
    }
  }


  /**
   * liefert das aktuelle Datum als String im default-Format (ISO) zur?ck.
   * 
   * @exception Exception wird ausgel?st, falls ein Fehler vorliegt.
   */
  public static String getDateAsString() throws Exception {

    SimpleDateFormat formatter = new SimpleDateFormat(SOSDate.dateFormat);
    return formatter.format(SOSDate.getDate());
  }


  /**
   * liefert das im ISO-Format angegebene Datum als Date zur?ck.
   * 
   * @param dateStr Datum im ISO-Format yyyy-mm-dd
   * @exception Exception wird ausgel?st, falls ein Fehler vorliegt.
   */
  public static Date getDate(String dateStr) throws Exception {

    SimpleDateFormat formatter;

    try {
      formatter = new SimpleDateFormat(SOSDate.dateFormat);
    } catch (Exception e) {
      throw (new Exception("invalid date format string: " + e.toString()));
    }

    try {
      return formatter.parse(dateStr);
    } catch (Exception e) {
      throw (new Exception("illegal date value: " + e.toString()));
    }
  }


  /**
   * liefert das angegebene Datum als String im ISO-Format zur?ck.
   * 
   * @param date Datum vom Typ Date
   * @exception Exception wird ausgel?st, falls ein Fehler vorliegt.
   */
  public static String getDateAsString(Date date) throws Exception {

    SimpleDateFormat formatter = new SimpleDateFormat(SOSDate.dateFormat);
    return formatter.format(date);
  }


  /**
   * liefert das im angegebenen Format angegebene Datum als Date zur?ck.
   * 
   * @param dateStr Datum im ISO-Format yyyy-mm-dd bzw. dem angegebenen Format
   * @param dateFormat Format-String, z.B. yyyy-mm-dd
   * @exception Exception wird ausgel?st, falls ein Fehler vorliegt.
   */
  public static Date getDate(String dateStr, String dateFormat)
      throws Exception {

    SimpleDateFormat formatter;

    try {
      formatter = new SimpleDateFormat(dateFormat);
    } catch (Exception e) {
      throw (new Exception("invalid date format string: " + e.toString()));
    }

    try {
      return formatter.parse(dateStr);
    } catch (Exception e) {
      throw (new Exception("illegal date string: " + e.toString()));
    }
  }


  /**
   * liefert das angegebene Datum vom Typ Date als String im angegebenen Format
   * zur?ck.
   * 
   * @param date Datum vom Typ Date
   * @param dateFormat Format-String, z.B. yyyy-mm-dd
   * @exception Exception wird ausgel?st, falls ein Fehler vorliegt.
   */
  public static String getDateAsString(Date date, String dateFormat)
      throws Exception {

    SimpleDateFormat formatter = new SimpleDateFormat(dateFormat);
    return formatter.format(date);
  }


  /**
   * liefert die aktuelle Zeit als Date zur?ck.
   * 
   * @exception Exception wird ausgel?st, falls ein Fehler vorliegt.
   */
  public static Date getTime() throws Exception {

    try {
      Calendar now = Calendar.getInstance();
      return now.getTime();
    } catch (Exception e) {
      throw (new Exception("illegal date value: " + e.toString()));
    }

  }


  /**
   * liefert die aktuelle Zeit als String im default-Format zur?ck.
   * 
   * @exception Exception wird ausgel?st, falls ein Fehler vorliegt.
   */
  public static String getTimeAsString() throws Exception {

    SimpleDateFormat formatter = new SimpleDateFormat(SOSDate.dateTimeFormat);
    return formatter.format(SOSDate.getTime());
  }


  /**
   * liefert die im ISO-Format angegebene Zeit als Date zur?ck.
   * 
   * @param dateTimeStr Zeit im ISO-Format yyyy-mm-dd hh:mm:ss
   * @exception Exception wird ausgel?st, falls ein Fehler vorliegt.
   */
  public static Date getTime(String dateTimeStr) throws Exception {

    SimpleDateFormat formatter;

    try {
      formatter = new SimpleDateFormat(SOSDate.dateTimeFormat);
    } catch (Exception e) {
      throw (new Exception("invalid date format string: " + e.toString()));
    }

    try {
      Calendar now = Calendar.getInstance();
      return formatter.parse(dateTimeStr);
    } catch (Exception e) {
      throw (new Exception("illegal date value: " + e.toString()));
    }

  }


  /**
   * liefert die angegebene Zeit als String im default-Format zur?ck.
   * 
   * @param date Zeit vom Typ Date
   * @exception Exception wird ausgel?st, falls ein Fehler vorliegt.
   */
  public static String getTimeAsString(Date date) throws Exception {

    SimpleDateFormat formatter = new SimpleDateFormat(SOSDate.dateTimeFormat);
    return formatter.format(date);
  }


  /**
   * liefert die angegebene Zeit im angegebenen Format als Date zur?ck.
   * 
   * @param dateTimeStr Zeit im ISO-Format yyyy-mm-dd hh:mm:ss bzw. dem
   *         angegebenen Format
   * @param dateTimeFormat Format-String, z.B. yyyy-mm-dd hh:mm:ss
   * @exception Exception wird ausgel?st, falls ein Fehler vorliegt.
   */
  public static Date getTime(String dateTimeStr, String dateTimeFormat)
      throws Exception {

    SimpleDateFormat formatter;

    try {
      formatter = new SimpleDateFormat(dateTimeFormat);
    } catch (Exception e) {
      throw (new Exception("invalid date format string: " + e.toString()));
    }

    try {
      return formatter.parse(dateTimeStr);
    } catch (Exception e) {
      throw (new Exception("illegal date value: " + e.toString()));
    }

  }


  /**
   * liefert die angegebene Zeit als String im angegebenen Format zur?ck.
   * 
   * @param date Zeit vom Typ Date
   * @param dateTimeFormat Format-String, z.B. yyyy-mm-dd hh:mm:ss
   * @exception Exception wird ausgel?st, falls ein Fehler vorliegt.
   */
  public static String getTimeAsString(Date date, String dateTimeFormat)
      throws Exception {

    SimpleDateFormat formatter = new SimpleDateFormat(dateTimeFormat);
    return formatter.format(date);
  }


  /**
   * Validiert ein Datum anhand der standard Formate
   * 
   * @param text Datum
   * @param dateStyle Format
   * @param locale Lokale
   * @return G&uuml;ltigkeit
   */
  public static boolean isValidDate(String text, int dateStyle, Locale locale) {
    try {
      getDate(text, dateStyle, locale);
      return true;
    } catch (Exception e) {
      return false;
    }
  }


  /**
   * Validiert eine Uhrzeit anhand der standard Formate
   * 
   * @param text Datum
   * @param locale Lokale
   * @return G&uuml;ltigkeit
   */
  public static boolean isValidTime(String text, int timeStyle, Locale locale) {
    try {
      getTime(text, timeStyle, locale);
      return true;
    } catch (Exception e) {
      return false;
    }
  }


  /**
   * Validiert eine Datum-Uhrzeit-Angabe anhand der standard Formate
   * 
   * @param text Datum
   * @param dateStyle Format
   * @param locale Lokale
   * @return G&uuml;ltigkeit
   */
  public static boolean isValidDateTime(String text, int dateStyle,
                                        int timeStyle, Locale locale) {
    try {
      getDateTime(text, dateStyle, timeStyle, locale);
      return true;
    } catch (Exception e) {
      return false;
    }
  }


  /**
   * Liefert eine lokale Datumdarstellung als Text anhand der standard Formate
   * 
   * @param date Datum vom Typ Date
   * @param dateStyle Format
   * @param locale Lokale
   * @return Datum als Text
   */
  public static String getDateAsString(Date date, int dateStyle, Locale locale) {
    DateFormat formatter = DateFormat.getDateInstance(dateStyle, locale);
    return formatter.format(date);
  }


  /**
   * Liefert eine lokale Zeitdarstellung als Text anhand der standard Formate
   * 
   * @param date Datum vom Typ Date
   * @param dateStyle Format
   * @param locale Lokale
   * @return Zeit als Text
   */
  public static String getTimeAsString(Date date, int timeStyle, Locale locale) {
    DateFormat formatter = DateFormat.getTimeInstance(timeStyle, locale);
    return formatter.format(date);
  }


  /**
   * Liefert eine lokale Datum-Zeit-Darstellung als Text anhand der standard
   * Formate
   * 
   * @param date Datum vom Typ Date
   * @param dateStyle Format
   * @param locale Lokale
   * @return Datum-Zeit als Text
   */
  public static String getDateTimeAsString(Date date, int dateStyle,
                                           int timeStyle, Locale locale) {
      
    DateFormat formatter = DateFormat.getDateTimeInstance(dateStyle, timeStyle,locale);
    
    return formatter.format(date);
  }

  
  /**
   * Liefert eine lokale Datum-Zeit-Darstellung als Text anhand des SOSDate.outputDateTimeFormat
   * 
   * @param datestr		Datum als String im SOSDate.dateTimeFormat format oder %now f�r aktuelles Datum
   * @return
   * @throws Exception
   */
  public static String getDateTimeAsString(String datestr) throws Exception {
     
      return getDateTimeAsString(datestr,null);
  }
  
  /**
   * Liefert eine lokale Datum-Zeit-Darstellung als Text
   * 
   * @param datestr					Datum als String im SOSDate.dateTimeFormat format oder %now f�r aktuelles Datum
   * @param outputDateTimeFormat	Ausgabe Format<br>
   * 								zB: dd.MM.yyyy HH:mm:ss,dd.MM.yy HH:mm ...
   * 									MM/dd/yy HH:mm ...
   * 								Die Reihenfolge und Delimiter bei den formatierungs Elementen sind belibig,<br>
   * 								es m�ssen nur g�ltige Java Datum/Zeit Format-Elemente sein
   * @return
   * @throws Exception
   */
  public static String getDateTimeAsString(String datestr, String outputDateTimeFormat ) throws Exception {
    
    Date date = null;
    
    if(datestr.equals("%now")){
        date = new Date();
    }
    else{
        date = SOSDate.getTime(datestr);
    }
    
    if(outputDateTimeFormat == null || outputDateTimeFormat.length() == 0){
        outputDateTimeFormat = SOSDate.getOutputDateTimeFormat();
    }
 
    DateFormat formatter = new SimpleDateFormat(outputDateTimeFormat);
    
  return formatter.format(date);
  }

  
  /**
   * Liefert eine lokale Datum-Zeit-Darstellung als Text
   * 
   * @param datestr					Datum als String im SOSDate.dateTimeFormat format oder %now f�r aktuelles Datum
   * @param outputDateTimeFormat	Ausgabe Format<br>
   * 								zB: dd.MM.yyyy HH:mm:ss,dd.MM.yy HH:mm ...
   * 									MM/dd/yy HH:mm ...
   * 								Die Reihenfolge und Delimiter bei den formatierungs Elementen sind belibig,<br>
   * 								es m�ssen nur g�ltige Java Datum/Zeit Format-Elemente sein
   * @return
   * @throws Exception
   */
  public static String getDateTimeAsString(Date date, String outputDateTimeFormat ) throws Exception {
    
    if(outputDateTimeFormat == null || outputDateTimeFormat.length() == 0){
        outputDateTimeFormat = SOSDate.getOutputDateTimeFormat();
    }
 
    DateFormat formatter = new SimpleDateFormat(outputDateTimeFormat);
    
  return formatter.format(date);
  }
  

  /**
   * Erstellt aus einer lokalen Datumdarstellung ein Date-Objekt anhand von
   * standard Formaten
   * 
   * @param text textuelle Darstellung des Datums
   * @param dateStyle Format
   * @param locale Lokale
   * @return Datum vom Typ Date
   * @throws ParseException
   */
  public static Date getDate(String text, int dateStyle, Locale locale)
      throws ParseException {
    DateFormat formatter = DateFormat.getDateInstance(dateStyle, locale);
    return formatter.parse(text);
  }


  /**
   * Erstellt aus einer lokalen Zeitdarstellung ein Date-Objekt anhand von
   * standard Formaten
   * 
   * @param text textuelle Darstellung der Zeit
   * @param dateStyle Format
   * @param locale Lokale
   * @return Datum vom Typ Date
   * @throws ParseException
   */
  public static Date getTime(String text, int timeStyle, Locale locale)
      throws ParseException {
    DateFormat formatter = DateFormat.getTimeInstance(timeStyle, locale);
    return formatter.parse(text);
  }


  /**
   * Erstellt aus einer lokalen Datum-Zeit-Darstellung ein Date-Objekt anhand
   * von standard Formaten
   * 
   * @param text textuelle Darstellung von Datum-Zeit
   * @param dateStyle Format
   * @param locale Lokale
   * @return Datum vom Typ Date
   * @throws ParseException
   */
  public static Date getDateTime(String text, int dateStyle, int timeStyle,
                                 Locale locale) throws ParseException {
    DateFormat formatter = DateFormat.getDateTimeInstance(dateStyle, timeStyle,
        locale);
    return formatter.parse(text);
  }


  /**
   * Liefert den lokalen Pattern zum Datum von einem standard Format
   * 
   * @param dateStyle Format
   * @param locale Lokale
   * @return lokaler Pattern
   */
  public static String getDatePattern(int dateStyle, Locale locale) {
    SimpleDateFormat formatter = (SimpleDateFormat) DateFormat.getDateInstance(
        dateStyle, locale);
    return formatter.toLocalizedPattern();
  }


  /**
   * Liefert den lokalen Pattern zur Zeit von einem standard Format
   * 
   * @param dateStyle Format
   * @param locale Lokale
   * @return lokaler Pattern
   */
  public static String getTimePattern(int timeStyle, Locale locale) {
    SimpleDateFormat formatter = (SimpleDateFormat) DateFormat.getTimeInstance(
        timeStyle, locale);
    return formatter.toLocalizedPattern();
  }


  /**
   * Liefert den lokalen Pattern zur Datum-Zeit von einem standard Format
   * 
   * @param dateStyle Format
   * @param locale Lokale
   * @return lokaler Pattern
   */
  public static String getDateTimePattern(int dateStyle, int timeStyle,
                                          Locale locale) {
    SimpleDateFormat formatter = (SimpleDateFormat) DateFormat
        .getDateTimeInstance(dateStyle, timeStyle, locale);
    return formatter.toLocalizedPattern();
  }
  
  
  
  /**
   * Liefert default Ausgabe Format
   * 
   * @return Returns the outputDateTimeFormat.
   */
  public static String getOutputDateTimeFormat() {
      return outputDateTimeFormat;
  }
  /**
   * Setzt default Ausgabe Format
   * 
   * @param outputDateTimeFormat Ausgabe Format
   */
  public static void setOutputDateTimeFormat(String outputDateTimeFormat) {
      SOSDate.outputDateTimeFormat = outputDateTimeFormat;
  }
  
   /**
   * Methode erwarten einen String im SOSDate.dateFormat Format (default yyyy-MM-dd)<br>
   * und liefert sprachabhngig formatiertes Datum (aktuelle Locale : SOSDate.locale) als String zurck<br>
   * <br>
   * Ausgabeformat wird ber die Eigenschaft SOSDate.dateStyle bestimmt (default Jahr - 2.stellig) 
   * <br>
   * 
   * @param datestr Datum als String im SOSDate.dateFormat Format
   * @return		sprachabhngiges Datum als String
   * @throws Exception
   */
  public static String getLocaleDateAsString(String datestr) throws Exception {
  
      SimpleDateFormat sdf 	= new SimpleDateFormat(SOSDate.dateFormat);
      Date date 						= sdf.parse(datestr);
      DateFormat formatter 	= DateFormat.getDateInstance(SOSDate.dateStyle,SOSDate.locale);
       
   return formatter.format(date);
  }
  
  
  
   /**
   * Methode erwarten einen String im SOSDate.dateTimeFormat Format (default yyyy-MM-dd HH:mm:ss)<br>
   * und liefert sprachabhngig formatiertes Datum und Zeit (aktuelle Locale : SOSDate.locale) als String zurck<br>
   * <br>
   * Ausgabeformat wird ber die Eigenschaften SOSDate.dateStyle und SOSDate.timeStyle bestimmt<br>
   * (default Jahr - 2.stellig und die Zeit ohne Millisekunden)<br>
   * 
   * @param datestr Datum und Zeit als String im SOSDate.dateTimeFormat Format
   * @return		sprachabhngiges Datum und Zeit als String
   * @throws Exception
   */
  public static String getLocaleDateTimeAsString(String datestr) throws Exception {
  
      SimpleDateFormat sdf 	= new SimpleDateFormat(SOSDate.dateTimeFormat);
      Date date 						= sdf.parse(datestr);
      DateFormat formatter 	= DateFormat.getDateTimeInstance(SOSDate.dateStyle,SOSDate.timeStyle,SOSDate.locale);
  
    return formatter.format(date);
  }

   /**
   * Methode erwarten ein Date Objekt<br>
   * und liefert sprachabhngig formatiertes Datum (aktuelle Locale : SOSDate.locale) als String zurck<br>
   * <br>
   * Ausgabeformat wird ber die Eigenschaft SOSDate.dateStyle bestimmt (default Jahr - 2.stellig) 
   * <br>
   * 
   * @param datestr Datum als String im SOSDate.dateFormat Format
   * @return		sprachabhngiges Datum als String
   * @throws Exception
   */
  public static String getLocaleDateAsString(Date date) throws Exception {
  
      DateFormat formatter 	= DateFormat.getDateInstance(SOSDate.dateStyle,SOSDate.locale);
   return formatter.format(date);
  }
  
  
  
   /**
   * Methode erwarten ein Date Objekt<br>
   * und liefert sprachabhngig formatiertes Datum und Zeit (aktuelle Locale : SOSDate.locale) als String zurck<br>
   * <br>
   * Ausgabeformat wird ber die Eigenschaften SOSDate.dateStyle und SOSDate.timeStyle bestimmt<br>
   * (default Jahr - 2.stellig und die Zeit ohne Millisekunden)<br>
   * 
   * @param datestr Datum und Zeit als String im SOSDate.dateTimeFormat Format
   * @return		sprachabhngiges Datum und Zeit als String
   * @throws Exception
   */
  public static String getLocaleDateTimeAsString(Date date) throws Exception {
  
      DateFormat formatter 	= DateFormat.getDateTimeInstance(SOSDate.dateStyle,SOSDate.timeStyle,SOSDate.locale);
       
   return formatter.format(date);
  }
  
  /**
   * Formatiert ein GregorianCalendar Objekt in ISO-Format und liefert als String zurck
   * 
   * @param date
   * @return
   * @throws Exception
   */
  public static String getISODateTimeAsString(GregorianCalendar date) throws Exception{
      SimpleDateFormat iso_format = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
      return iso_format.format(date.getTime());
  }
 
  /**
   * Formatiert ein GregorianCalendar Objekt in ISO-Format und liefert als String zurck
   * 
   * @param date
   * @return
   * @throws Exception
   */
  public static String getISODateAsString(GregorianCalendar date) throws Exception{
      SimpleDateFormat iso_format = new SimpleDateFormat("yyyy-MM-dd");
      return iso_format.format(date.getTime());
  }
  
}