// $Id: series_factory.cxx 3546 2005-04-03 08:30:29Z jz $          Joacim Zschimmer

// Der Factory-Prozessor (Version 2)

#include "precomp.h"

#include <wchar.h>
#include <io.h>
#include <fcntl.h>
#include <strstream>

#include <objbase.h>        // OLE
#include "msscript.h"       // GUID fr Scripting Engine
#include <activscp.h>       // Scripting Engine

#include "../kram/sosstrng.h"
#include "../kram/sos.h"
#include "../file/anyfile.h"
#include "../kram/log.h"

#include "../hostole/variables.h"

#include "factory_processor.h"
#include "factory_text.h"
#include "factory_rtf.h"
#include "series_factory_script.h"

//-------------------------------------------------------------------------------------------

#ifdef SYSTEM_WIN
    inline int fdatasync( int fn )  { return _commit( fn ); }
#endif

//-------------------------------------------------------------------------------------------

namespace sos {
namespace factory {

//-----------------------------------------------------------------------------Series_factory
// Aufbereitung einer Serie von Dokumenten

struct Series_factory : Iseries_factory,
                        Sos_ole_object
{
    void*                       operator new                ( size_t size )                 { return sos_alloc( size, "Series_factory" ); }
    void                        operator delete             ( void* ptr )                   { sos_free( ptr ); }


                                Series_factory              ();
                               ~Series_factory              ();

    USE_SOS_OLE_OBJECT

    STDMETHODIMP                get_Order_filename          ( BSTR* s )                     { *s = SysAllocString_string( _order_filename );  return NOERROR; }
    STDMETHODIMP                put_Order_filename          ( BSTR s )                      { _order_filename = bstr_as_string(s); return NOERROR; }
    STDMETHODIMP                get_Template_script_language( BSTR* s )                     { return _processor->get_Language(s); return NOERROR; }
    STDMETHODIMP                put_Template_script_language( BSTR s )                      { return _processor->put_Language(s); return NOERROR; }
    STDMETHODIMP                get_Template_dir            ( BSTR* s )                     { return _processor->get_Template_dir(s); }
    STDMETHODIMP                put_Template_dir            ( BSTR s )                      { return _processor->put_Template_dir(s); }
    STDMETHODIMP                get_Template_filename       ( BSTR* s )                     { *s = SysAllocString_string( _template_filename ); return NOERROR; }
    STDMETHODIMP                put_Template_filename       ( BSTR );
    STDMETHODIMP                get_Head_filename           ( BSTR* s )                     { *s = SysAllocString_string( _head_filename ); return NOERROR; }
    STDMETHODIMP                put_Head_filename           ( BSTR s )                      { _head_filename = bstr_as_string(s); return NOERROR; }
    STDMETHODIMP                get_Start_script            ( BSTR* s )                     { *s= SysAllocString( _start_script_bstr ); return NOERROR; }
    STDMETHODIMP                put_Start_script            ( BSTR start_script )           { _start_script_bstr = start_script; return NOERROR; }
    STDMETHODIMP                get_Document_dir            ( BSTR* s )                     { return _processor->get_Document_dir(s); }
    STDMETHODIMP                put_Document_dir            ( BSTR s )                      { return _processor->put_Document_dir(s); }
    STDMETHODIMP                get_On_error_continue       ( VARIANT_BOOL* b )             { *b = _on_error_continue; return NOERROR; }
    STDMETHODIMP                put_On_error_continue       ( VARIANT_BOOL b )              { _on_error_continue = b != 0; return NOERROR; }
    STDMETHODIMP                get_Collect                 ( int* collect )                { return _processor->get_Collect( collect ); return NOERROR; }
    STDMETHODIMP                put_Collect                 ( int collect )                 { return _processor->put_Collect( collect ); return NOERROR; }
    STDMETHODIMP                get_Context                 ( VARIANT* v )                  { *v = _context;  return NOERROR; }
    STDMETHODIMP                put_Context                 ( VARIANT v )                   { _context = v;  return NOERROR; }
    STDMETHODIMP                put_Processing_object_constructor( IDispatch* ); 
    STDMETHODIMP                put_Output_object_constructor( IDispatch* i )               { _output_object_constructor = i; return ERROR; }


    STDMETHODIMP                get_Rerun_filename          ( BSTR* s )                     { *s = SysAllocString_string( _rerun_filename ); return NOERROR; }
    STDMETHODIMP                put_Rerun_filename          ( BSTR s )                      { _rerun_filename = bstr_as_string(s); return NOERROR; }
    STDMETHODIMP                get_Document_filename       ( BSTR* filename )              { *filename = SysAllocString_string( _document_filename ); return NOERROR; }
    STDMETHODIMP                put_Document_filename       ( BSTR );
    STDMETHODIMP                get_Processor               ( Ifactory_processor** p )      { *p = _processor; if(*p) (*p)->AddRef(); return NOERROR; }
    STDMETHODIMP                get_Record_nr               ( int* i )                      { *i = _record_nr; return NOERROR; }
    STDMETHODIMP                get_Rerunning               ( VARIANT_BOOL* b )             { *b = _rerun; return NOERROR; }


    STDMETHODIMP                Open                        ();
    STDMETHODIMP                Close                       ();
    STDMETHODIMP                Process                     ( VARIANT_BOOL* );
    STDMETHODIMP                Process_all                 ();
    STDMETHODIMP                Eof                         ( VARIANT_BOOL* );


    void                        clean_up                    ( bool finished = false );
    void                        get_record                  ();
    void                        begin                       ();
    void                        prepare                     ();
    bool                        process_document            ();
    int                         get_char_from_rerun_file    ();
    int                         get_int_from_rerun_file     ();
    bool                        set_next_rerun_record_nr    ();
    void                        end_reading_rerun_file      ();
    void                        begin_writing_rerun_file    ();
    void                        write_rerun_number          ( int next_record_nr );
    void                        check_document_written      ( int next_record_nr );
    void                        write_single_rerun_number   ( int record_nr );

    Fill_zero                  _zero_;
    string                     _order_filename;
    string                     _template_filename;
    string                     _head_filename;
    Bstr                       _start_script_bstr;
    bool                       _on_error_continue;
    Any_file                   _order_file;
    string                     _document_filename;
    string                     _rerun_filename;
    FILE*                      _rerun_file;
    int                        _rerun_pos;                 
    bool                       _document_has_to_be_written;
    int                        _rerun_record_nr_written;
    int                        _record_nr;
    bool                       _eof;
    int                        _begun;
    bool                       _rerun;                      // _rerun_file wird gelesen
    int                        _rerun_high;                 // Nummer des Satzes, bis zu dem aufbereitet werden soll
    bool                       _rerun_file_was_filled;
    bool                       _error_in_rerun;             // Bei einem Fehler im Wiederanlauf die Wiederanlaufdatei unverndert lassen
    Variant                    _context;
    ptr<Processor>             _processor; 
    ptr<Hostware_variables>    _parameters;
    ptr<IDispatch>             _processing_object_constructor;
    ptr<IDispatch>             _processing_object;
    ptr<IDispatch>             _output_object_constructor;
    ptr<IDispatch>             _output_object;
};

//-------------------------------------------------------------------------------Typbibliothek

DESCRIBE_CLASS_CREATABLE( &hostole_typelib, Series_factory, series_factory, CLSID_Series_factory, "hostWare.Series_factory", "1.0" );

//--------------------------------------------------------------Series_factory::Series_factory

Series_factory::Series_factory()                              
: 
    Sos_ole_object( series_factory_class_ptr, this ), 
    _zero_(this+1)
{
    _processor = new Processor;
}

//-------------------------------------------------------------Series_factory::~Series_factory

Series_factory::~Series_factory()
{
    _processor = NULL;
    if( _rerun_file )  fclose( _rerun_file ), _rerun_file = NULL;
    //_script.close();
}

//-------------------------------------------------------Series_factory::put_template_filename

STDMETHODIMP Series_factory::put_Template_filename( BSTR s )
{ 
    _template_filename = bstr_as_string( s );
    return NOERROR; 
}

//-------------------------------------------------------Series_factory::put_document_filename

STDMETHODIMP Series_factory::put_Document_filename( BSTR filename_bstr )
{ 
    _document_filename = bstr_as_string( filename_bstr );

    return _processor->put_Document_filename( filename_bstr ); 
}

//-------------------------------------------Series_factory::put_processing_object_constructor

STDMETHODIMP Series_factory::put_Processing_object_constructor( IDispatch* constructor )
{
    _processing_object_constructor = constructor;
    return ERROR;
}

//------------------------------------------------------------------------Series_factory::Open

STDMETHODIMP Series_factory::Open()
{
    HRESULT         hr = NOERROR;
    Dynamic_area    record;

    try 
    {
        _order_file.open( _order_filename, Any_file::in );

        if( !_rerun_filename.empty() )
        {
            if( _rerun_file )  fclose( _rerun_file );

            _rerun_file = fopen( _rerun_filename.c_str(), "r" ); 
            if( !_rerun_file  &&  errno != ENOENT )  throw_errno( errno, _rerun_filename.c_str() );

            if( _rerun_file )  
            {
                _rerun = true;
                _rerun_file_was_filled = false;    // Bis zum Beweis des Gegenteils
                LOG( "Series_factory Wiederanlauf\n" );
            }
            else
                begin_writing_rerun_file();
        }
    }
    catch( const Xc&         x )  { hr = _set_excepinfo( x, "hostWare.Series_factory::open" ); }
    catch( const exception&  x )  { hr = _set_excepinfo( x, "hostWare.Series_factory::open" ); }
    catch( const _com_error& x )  { hr = _set_excepinfo( x, "hostWare.Series_factory::open" ); }
                                                         
    return hr;
}

//-----------------------------------------------------------------------Series_factory::Close

STDMETHODIMP Series_factory::Close()
{
    HRESULT hr = NOERROR;
    HRESULT close_hr = NOERROR;

    try 
    {
        if( _processor )  
        {
            hr = _processor->Close_output_file();  
            if( !FAILED(hr) )  check_document_written( _record_nr ); 

            close_hr = _processor->Close();
            _processor = NULL;
        }

        if( _rerun_file )  fclose( _rerun_file ), _rerun_file = NULL;

        _order_file.close();

        if( FAILED(hr) )  throw_ole( hr, "hostWare.Series_factory::close" );
        hr = close_hr;
    }
    catch( const Xc&         x )  { hr = _set_excepinfo( x, "hostWare.Series_factory::close" ); }
    catch( const exception&  x )  { hr = _set_excepinfo( x, "hostWare.Series_factory::close" ); }
    catch( const _com_error& x )  { hr = _set_excepinfo( x, "hostWare.Series_factory::close" ); }

    return hr;
}

//--------------------------------------------------------------------Series_factory::clean_up

void Series_factory::clean_up( bool finished )
{
    _processing_object = NULL;
    _parameters = NULL;
    _begun = false;

    if( finished && _rerun_file )
    {
        fclose( _rerun_file );
        _rerun_file = NULL;
        unlink( _rerun_filename.c_str() );
    }
}

//----------------------------------------------------Series_factory::get_char_from_rerun_file

int Series_factory::get_char_from_rerun_file()
{
    int c = fgetc( _rerun_file );
    
    while( c == ' ' || c == '\t' )  c = get_char_from_rerun_file();      // Abweichende Gestalt tolerieren

    if( c == EOF )  if( feof(_rerun_file) || errno == 0 )  ; //throw_eof_error();
                                                     else  throw_errno( errno, c_str( _rerun_filename ) );

    return c;
}

//-----------------------------------------------------Series_factory::get_int_from_rerun_file

int Series_factory::get_int_from_rerun_file()
{
    int n;
    int c = get_char_from_rerun_file();

    while( isspace( c ) )  c = get_char_from_rerun_file();      // Abweichende Gestalt tolerieren

    if( !isdigit(c) )  throw_xc( "FACTORY-303", "Ziffer erwartet" );
    
    _rerun_file_was_filled = true;

    n = c - '0';
    
    while(1)
    {
        c = get_char_from_rerun_file();
        if( !isdigit(c) )  { ungetc( c, _rerun_file ); return n; }
        n = 10*n + c - '0';
    }
}

//----------------------------------------------------Series_factory::set_next_rerun_record_nr

bool Series_factory::set_next_rerun_record_nr()
{
    int from_recnr = get_int_from_rerun_file();

    int c = get_char_from_rerun_file();

    if( c == '-' ) 
    {
        c = get_char_from_rerun_file();
        if( c == '\r'  ||  c == '\n'  ||  c == EOF )
        {                                       // "VONNR-"
            ungetc( c, _rerun_file );
            _rerun_high = INT_MAX;
            c = EOF;
            goto SKIP;
        }
        else                                    // "VONNR-BISNR"
        {
            ungetc( c, _rerun_file );
            _rerun_high = get_int_from_rerun_file();
        }
    }
    else
    {
        _rerun_high = from_recnr;               // "NR"
        ungetc( c, _rerun_file );
    }

    c = get_char_from_rerun_file();
    if( c == '\r' )  c = get_char_from_rerun_file();
    if( c != '\n' && c != EOF )  throw_xc( "FACTORY-303", "Zeilenende erwartet" );

    while( isspace( c ) )  c = get_char_from_rerun_file();      // Toleranz
    if( c != EOF )  ungetc( c, _rerun_file );

SKIP:        
    while( _record_nr + 1 < from_recnr )  _record_nr++, get_record();

    if( c == EOF )  
    {
        end_reading_rerun_file();
        if( !_rerun_file_was_filled )  _rerun_high = INT_MAX;
    }

    return true;
}

//------------------------------------------------------Series_factory::end_reading_rerun_file

void Series_factory::end_reading_rerun_file()
{
    LOG( "Series_factory: Wiederanlaufdatei ist abgearbeitet\n" );

    fclose( _rerun_file );
    _rerun_file = NULL;

    _rerun = false;

    if( !_error_in_rerun )  begin_writing_rerun_file();
}

//----------------------------------------------------Series_factory::begin_writing_rerun_file

void Series_factory::begin_writing_rerun_file()
{
    _rerun_file = fopen( _rerun_filename.c_str(), "w" );
    if( !_rerun_file )  throw_errno( errno, _rerun_filename.c_str() );

    write_rerun_number( _record_nr + 1 );
}

//----------------------------------------------------------Series_factory::write_rerun_number

void Series_factory::write_rerun_number( int next_record_nr )
{
    int ret;

    if( next_record_nr == _rerun_record_nr_written )  return;

    ret = fseek( _rerun_file, _rerun_pos, SEEK_SET );       if( ret == -1 )  throw_errno( errno, _rerun_filename.c_str() );
    ret = fprintf( _rerun_file, "%d-\n", next_record_nr );  if( ret == -1 )  throw_errno( errno, _rerun_filename.c_str() );
    ret = fflush( _rerun_file );                            if( ret == -1 )  throw_errno( errno, _rerun_filename.c_str() );
    ret = fdatasync( fileno( _rerun_file ) );               if( ret == -1 )  throw_errno( errno, _rerun_filename.c_str() );

    _rerun_record_nr_written = next_record_nr;
}

//----------------------------------------------------Series_factory::write_single_rerun_number

void Series_factory::write_single_rerun_number( int record_nr )
{
    int ret;

    ret = fseek( _rerun_file, _rerun_pos, SEEK_SET );       if( ret == -1 )  throw_errno( errno, _rerun_filename.c_str() );
    ret = fprintf( _rerun_file, "%d\n", record_nr );        if( ret == -1 )  throw_errno( errno, _rerun_filename.c_str() );

    _rerun_pos = ftell( _rerun_file );

    if( _record_nr + 1 < _rerun_high )  write_rerun_number( record_nr + 1 );
}

//------------------------------------------------------------------Series_factory::get_record

void Series_factory::get_record()
{
    HRESULT         hr;
    Dynamic_area    record;
    Dynamic_area    hilfspuffer;

    if( _order_filename.empty() )
    {
        if( !_processing_object )  throw_xc( "FACTORY-305", "order_filename" );

        LOG( "factory_get_parameters()\n" );
        Variant v = com_call( _processing_object, "factory_get_parameters" );

        if( v.vt == VT_EMPTY    )  throw_eof_error();
        if( v.vt == VT_NULL     )  throw_eof_error();
        if( v.vt != VT_DISPATCH )  throw_xc( "FACTORY-302" );
        if( v.pdispVal == NULL  )  throw_eof_error();
        hr = v.pdispVal->QueryInterface( IID_Variables, _parameters.void_pp() );
        if( FAILED(hr) )  throw_xc( "FACTORY-302" );
    }
    else
    {
        _parameters = new Hostware_variables;

        _order_file.get( &record );

        Sos_ptr<Record_type> record_type = _order_file.spec()._field_type_ptr;

        for( int i = 0; i < record_type->field_count(); i++ )
        {
            Field_descr* f = record_type->field_descr_ptr(i);
            Bstr         name_bstr = f->name();
            Variant      v;

            field_to_variant( f, record.byte_ptr(), &v, &hilfspuffer, (LCID)0 );

            if( SysStringLen( name_bstr ) > 0 )  _parameters->put_Value( name_bstr, &v );
                                           else  LOG( "Leerer Feldname wird ignoriert. Wert=" << debug_string_from_variant(v) << '\n' );
        }
    }
}

//-----------------------------------------------------------------------Series_factory::begin

void Series_factory::begin()
{
    HRESULT hr;

    if( _eof )  throw_eof_error();
    if( _begun )  return;
    _begun = true;

    if( _rerun  &&  _record_nr >= _rerun_high )
    {
        bool ok = set_next_rerun_record_nr();
        if( !ok )  throw_eof_error();
    }

    _processing_object = NULL;  // Vorsichtshalber

    _record_nr++;

    if( _processing_object_constructor )  
    {
        LOG( "processing_object_constructor()\n" );
        Variant v = com_call( _processing_object_constructor, "()", _context );
        hr = v.ChangeType( VT_DISPATCH );  if( FAILED(hr) )  throw_ole( hr, "processing_object_constructor()" );
        _processing_object = v.pdispVal;

        if( com_name_exists( _processing_object, "factory_processor" ) )
            com_property_putref( _processing_object, "factory_processor", Variant( _processor ) );
    }

    try
    {
        get_record();
    }
    catch( const Eof_error& ) 
    { 
        hr = _processor->Close_output_file();  if( FAILED(hr) )  throw_ole( hr, "hostWare.Factory_processor::close_output_file" );
        check_document_written( _record_nr ); 
        throw; 
    }
}

//---------------------------------------------------------------------Series_factory::prepare

void Series_factory::prepare()
{
    HRESULT hr = NOERROR;
    Variant v;
    string  template_filename;
    string  document_filename;
    string  head_filename;

    if( _processing_object )
    {
        v.Clear();
        v.vt = VT_DISPATCH | VT_BYREF;
        v.ppdispVal = (IDispatch**)&_parameters._ptr;
        LOG( "factory_prepare_parameters()\n" );
        com_call_if_exists( _processing_object, "factory_prepare_parameters", v );
    }

    hr = _processor->putref_Parameters( _parameters );                          if( FAILED(hr) )  throw_ole( hr, "hostWare.Factory_processor::parameters" );


    if( _processing_object )  
    {
        LOG( "factory_prepare_process()\n" );
        com_call_if_exists( _processing_object, "factory_prepare_process" );
    }

    v.Clear();
    LOG( "parse start_script\n" );
    hr = _processor->Parse( _start_script_bstr, ::scripttext_isvisible, &v );   if( FAILED(hr) )  throw_ole( hr, "hostWare.Factory_processor::parse" );


    v = _parameters->get_variant( L"factory_template_filename" );
    if( v.vt != VT_EMPTY ) {
        hr = v.ChangeType( VT_BSTR );                                           if( FAILED(hr) )  throw_ole( hr, "Parameter factory_template_filename ist kein String" );
        template_filename = bstr_as_string( v.bstrVal );
    }
    if( template_filename.empty() ) {
        if( !template_filename.empty() )  throw_xc( "FACTORY-301", "factory_template_filename" );
        template_filename = _template_filename;
    }
    hr = _processor->put_Template_filename( Bstr( template_filename.c_str() ) );  if( FAILED(hr) )  throw_ole( hr, "hostWare.Factory_processor::template_filename" );


    v = _parameters->get_variant( L"factory_head_filename" );
    if( v.vt != VT_EMPTY ) {
        hr = v.ChangeType( VT_BSTR );                                           if( FAILED(hr) )  throw_ole( hr, "Parameter factory_head_filename ist kein String" );
        head_filename = bstr_as_string( v.bstrVal );
    }
    if( head_filename.empty() ){
        head_filename = _head_filename;
    }
    hr = _processor->put_Head_filename( Bstr( head_filename.c_str() ) );        if( FAILED(hr) )  throw_ole( hr, "hostWare.Factory_processor::head_filename" );


    v = _parameters->get_variant( L"factory_document_filename" );
    if( v.vt != VT_EMPTY ) {
        hr = v.ChangeType( VT_BSTR );                                           if( FAILED(hr) )  throw_ole( hr, "Parameter factory_document_filename ist kein String" );
        document_filename = bstr_as_string( v.bstrVal );
    } 
    if( document_filename.empty() ) {
        if( !document_filename.empty() )  throw_xc( "FACTORY-301", "factory_document_filename" );
        document_filename = _document_filename;
    }
    hr = _processor->put_Document_filename( Bstr( document_filename.c_str() ) );  if( FAILED(hr) )  throw_ole( hr, "hostWare.Factory_processor::document_filename" );
}

//------------------------------------------------------Series_factory::check_document_written

void Series_factory::check_document_written( int next_record_nr )
{
    if( _document_has_to_be_written  &&  _processor->collected_documents_count() == 0 )
    {
        if( _output_object )
        {
            if( com_name_exists( _output_object, "factory_output_written" ) )
            {
                LOG( "factory_output_written()\n" );
                com_call( _output_object, "factory_output_written", Variant( Bstr( _processor->real_document_filename().c_str() ) ), 
                                                                    Variant( _processor->last_collected_documents_count() ) );
            }
            _output_object = NULL;
        }

        if( _rerun_file  &&  !_rerun )  write_rerun_number( next_record_nr );

        _document_has_to_be_written = false;
    }
}

//------------------------------------------------------------Series_factory::process_document

bool Series_factory::process_document()
{
    bool result;

    try
    {
        _processor->process();
        _document_has_to_be_written = true;
        
        if( _processing_object )  
        {
            LOG( "factory_on_success()\n" );
            com_call_if_exists( _processing_object, "factory_on_success" );
        }

        result = true;
    }
    catch( const exception& x )  
    { 
        _error_in_rerun |= _rerun;

        if( !_processing_object  ||  !com_name_exists( _processing_object, "factory_on_error" ) )  throw;

        if( _rerun_file && !_rerun )  write_single_rerun_number( _record_nr );

        LOG( "factory_on_error()\n" );
        com_call_if_exists( _processing_object, "factory_on_error", Variant( Bstr( x.what() ) ) );
        result = _on_error_continue;
    }

    if( _processing_object )  
    {
        LOG( "factory_after_process()\n" );
        com_call_if_exists( _processing_object, "factory_after_process" );
    }

    return result;
}

//---------------------------------------------------------------------....Series_factory::eof

STDMETHODIMP Series_factory::Eof( VARIANT_BOOL* eof )
{
    HRESULT hr = NOERROR;

    if( _eof )
    {
        *eof = true;
    }
    else
    {
        try 
        {
            begin();
            *eof = false;
        }
        catch( const Eof_error&    )  { *eof = true; }
        catch( const Xc&         x )  { hr = _set_excepinfo( x, "hostWare.Series_factory::eof" ); }
        catch( const exception&  x )  { hr = _set_excepinfo( x, "hostWare.Series_factory::eof" ); }
        catch( const _com_error& x )  { hr = _set_excepinfo( x, "hostWare.Series_factory::eof" ); }
    }

    return hr;
}

//---------------------------------------------------------------------Series_factory::process

STDMETHODIMP Series_factory::Process( VARIANT_BOOL* result )
{
    *result = false;

    HRESULT hr = NOERROR;
    bool    finished = false;

    try 
    {
        LOGI( "Series_factory::process record_nr=" << _record_nr << '\n' );

        try
        {
            begin();
        }
        catch( const Eof_error& )  { _eof = true; finished = true; goto ENDE; }


        prepare();

        check_document_written( _record_nr );   // Wenn sich der Vorlagendateiname etc. gendert hat, wird das erst bei prepare() bemerkt. Ausgabe ist dann abgeschlossen.

        if( !_output_object  &&  _output_object_constructor )  
        {
            LOG( "output_object_constructor()\n" );
            Variant v = com_call( _output_object_constructor, "()", _context );
            hr = v.ChangeType( VT_DISPATCH );  if( FAILED(hr) )  throw_ole( hr, "output_object_constructor()" );
            _output_object = v.pdispVal;
        }

        *result = process_document();

        _processing_object = NULL;
        _parameters = NULL;

        check_document_written( _record_nr + 1 );

      ENDE:
        clean_up( finished );
    }
    catch( const Xc&         x )  { clean_up();  hr = _set_excepinfo( x, "hostWare.Series_factory::process" ); }
    catch( const exception&  x )  { clean_up();  hr = _set_excepinfo( x, "hostWare.Series_factory::process" ); }
    catch( const _com_error& x )  { clean_up();  hr = _set_excepinfo( x, "hostWare.Series_factory::process" ); }

    return hr;
}

//-----------------------------------------------------------------Series_factory::process_all

STDMETHODIMP Series_factory::Process_all()
{
    HRESULT         hr = NOERROR;
    VARIANT_BOOL    ok = false;
   
    
    hr = Open();                    if( FAILED(hr) )  return hr;

    while(1)
    {
        hr = Process( &ok );        if( FAILED(hr) )  return hr;
        if( !ok )  break;
    }

    hr = Close();                   if( FAILED(hr) )  return hr;

    return hr;
}

//-------------------------------------------------------------------------------------------

} //namespace factory
} //namespace sos
