// $Id: perl_com.cxx 3546 2005-04-03 08:30:29Z jz $     Joacim Zschimmer

#include "zschimmer.h"
#include "z_com.h"
#include "log.h"
#include "scripting_engine.h"
#include "perl.h"
#include "perl_com.h"

using namespace std;
using namespace zschimmer::com;


namespace zschimmer {

//--------------------------------------------------------------------------------------perl_com_pm

extern const char perl_com_pm[] =
    "{ "
        "package " Z_PERL_IDISPATCH_PACKAGE_NAME "; "
      //"require 5.005_62; "
        "our $site;"
        "bootstrap " Z_PERL_IDISPATCH_PACKAGE_NAME ";"

        "sub new { shift; my $self = bless {}; $self->{__idispatch} = shift; return $self; } "

        "sub DESTROY "
        "{"
            "my $self = shift; "
            "if( defined $self->{__idispatch} ) { __destroy( $self->{__idispatch} ); undef $self->{__idispatch}; }"
        "}"

        "sub AUTOLOAD "
        "{"
            "my $self = shift; "
            "my $method_name = $AUTOLOAD; "
            "$method_name =~ s/^.*:://; "
            "if( substr($method_name,0,2) eq '__' ) {die;}"
            "$self->{__idispatch} = __resolve_name( $site, $self->{__name} )  unless defined $self->{__idispatch}; " 
            "__call( $self->{__idispatch}, 3, $method_name, @_ ); "
        "}"

        "sub LetProperty "            // Speicherleck: 80 Bytes laufen aus, wenn Perl beendet wird (use_engine="job").
        "{"
            "my $self = shift; "
            "my $property_name = shift; "
            "$self->{__idispatch} = __resolve_name( $site, $self->{__name} )  unless defined $self->{__idispatch}; " 
            "__call( $self->{__idispatch}, 4, $property_name, @_ ); "
        "}"

        "1;"
    "}";


template<class T>
inline void unused( const T& ) {}

//-------------------------------------------------------------------------------------bstr_from_pv

BSTR bstr_from_pv( SV* sv )
{
    STRLEN len = 0;
    const char* s = SvPV( sv, len );
    return bstr_from_string( s, len );
}

//-----------------------------------------------------------------------------XS_Zschimmer_destroy

XS( XS_Zschimmer_destroy )
{
    dXSARGS;

    bool    error = false;
    string  error_text;         // croak() verhindert den Aufruf des Destruktors.

    try
    {
        if( items != 1 )  throw_xc( "Z-PERL-100", "Usage: " Z_PERL_IDISPATCH_PACKAGE_NAME "::__destroy(idispatch)" );

        if( !SvIOK(ST(0)) )  throw_xc( "Z-PERL-100", Z_PERL_IDISPATCH_PACKAGE_NAME "::__destroy() 1. Parameter ist kein Integer" );

        IDispatch* idispatch = (IDispatch*)(int)SvIV(ST(0));

        if( idispatch )  
        {
            //fprintf( stderr, "perl_com.cxx: %x->Release()\n", (void*)idispatch );
            idispatch->Release();
        }
    }
    catch( const exception&  x ) { error = true;  error_text = x.what(); }
    catch( const _com_error& x ) { error = true;  error_text = string_from_bstr( x.Description() ); }

    if( error )  croak( error_text.c_str() );   // ACHTUNG: croak() kehrt nicht zurck!

    XSRETURN(1);
}

//------------------------------------------------------------------------XS_Zschimmer_resolve_name

XS( XS_Zschimmer_resolve_name )
{
    dXSARGS;

    bool    error = false;
    string  error_text;         // croak() verhindert den Aufruf des Destruktors.

    try
    {
        HRESULT    hr;
        IUnknown*  result = NULL;
        ITypeInfo* type_info;


        if( items != 2 )  throw_xc( "Z-PERL-100", "Usage: " Z_PERL_IDISPATCH_PACKAGE_NAME "::__resolve_name( site, name )" );

        if( !SvIOK(ST(0)) )  throw_xc( "Z-PERL-100", Z_PERL_IDISPATCH_PACKAGE_NAME "::__resolve_name() 1. Parameter ist kein Integer" );
        if( !SvPOK(ST(1)) )  throw_xc( "Z-PERL-100", Z_PERL_IDISPATCH_PACKAGE_NAME "::__resolve_name() 2. Parameter ist kein String" );

        IActiveScriptSite* site = (IActiveScriptSite*)(int)SvIV(ST(0));
        Bstr name_bstr;
        name_bstr.attach( bstr_from_pv( ST(1) ) );

        hr = site->GetItemInfo( name_bstr, SCRIPTINFO_IUNKNOWN, &result, &type_info );
        if( FAILED(hr) )  throw_com( hr, "GetItemInfo", string_from_bstr(name_bstr).c_str() );

        {
            int RETVAL;
            dXSTARG;
            RETVAL = (int)result;
            XSprePUSH; PUSHi((IV)RETVAL);
        }
    }
    catch( const exception&  x ) { error = true;  error_text = x.what(); }
    catch( const _com_error& x ) { error = true;  error_text = string_from_bstr( x.Description() ); }

    if( error )  croak( error_text.c_str() );   // ACHTUNG: croak() kehrt nicht zurck!

    XSRETURN(1);
}

//--------------------------------------------------------------------------------XS_Zschimmer_call

XS( XS_Zschimmer_call )
{
    dXSARGS;

    enum Result_type { rt_void, rt_int, rt_double, rt_string, rt_sv };
    Result_type result_type   = rt_void;
    
    IV          int_result    = 0;
    NV          double_result = 0;
    string      string_result;
    SV*         sv_result       = NULL;
    string      error_text;
    bool        error           = false;


    // *** Bei einem Fehler wird croak() gerufen. Diese Routine kehrt nicht zurck. Die Variablen string_result und error_text werden dann nicht freigeben. *** //

    try
    {
        HRESULT             hr;
        DISPID              dispid = 0;
        Dispparams          params;
        Variant             result;
        Excepinfo           excepinfo;
        uint                argnr = (uint)-1;
        string              method_name;
        const int           param_base = 3;


        if( items < param_base )  throw_xc( "Z-PERL-100", "Usage: " Z_PERL_IDISPATCH_PACKAGE_NAME "::__call( idispatch, flags, method, parameter, ... )" );

        if( !SvIOK(ST(0)) || !SvIOK(ST(1)) )  throw_xc( "Z-PERL-100", "Usage: " Z_PERL_IDISPATCH_PACKAGE_NAME "::__call( idispatch, flags, method, parameter, ... )" );

        IDispatch* idispatch = (IDispatch*)(int)SvIV(ST(0));

        if( !idispatch )  throw_xc( "Z-PERL-100", "NULL-Pointer" );

        int flags = (int)SvIV(ST(1));   // DISPATCH_METHOD, DISPATCH_PROPERTYPUT etc.

        if( SvIOK(ST(2)) )
        {
            dispid = (int)SvIV(ST(2));
        }
        else
        if( SvPOK(ST(2)) )  
        {
            method_name = string_from_pv( ST(2) );

            Bstr method_name_bstr = method_name;
            hr = idispatch->GetIDsOfNames( IID_NULL, &method_name_bstr, 1, STANDARD_LCID, &dispid );
            if( FAILED(hr) )  throw_com( hr, "GetIDsOfNames", method_name.c_str() );
        }
        else
            throw_xc( "Z-PERL-100", "Usage: " Z_PERL_IDISPATCH_PACKAGE_NAME "::__call( idispatch, method, parameter, ... )" );


        params.set_arg_count( items - param_base );

        for( int i = param_base; i < items; i++ )
        {
            SV* sv = ST(i);
            
            if( SvROK(sv) )  sv = SvRV(sv);                                   // Referenz?

            if( SvOK(sv) )                                                    // Definiert?
            {
                if( SvIOK(sv) )  params[i-param_base] = (int)SvIV(sv);        // Integer
                else
                if( SvNOK(sv) )  params[i-param_base] = (double)SvNV(sv);     // Double
                else
                if( SvPOK(sv) )                                               // String
                {
                    params[i-param_base].attach_bstr( bstr_from_pv(sv) );
                }
                else
                    throw_xc( "Z-PERL-100", as_string(i+1) + ". Parameter ist von unbekanntem Typ " + printf_string("%X",sv->sv_flags) );
            }
             
            //fprintf( stderr, "perl_com: %d. %s\n", i-param_base, string_from_variant(params[i-param_base]).c_str());
        }

        hr = idispatch->Invoke( dispid, IID_NULL, STANDARD_LCID, flags, &params, &result, &excepinfo, &argnr );
        if( FAILED(hr) ) 
        {
            if( hr == DISP_E_EXCEPTION )  throw_com_excepinfo( hr, &excepinfo, method_name.c_str(), "" );
        
            string a = method_name;
            if( ( hr == DISP_E_TYPEMISMATCH || hr == DISP_E_PARAMNOTFOUND )  &&  (int)argnr >= 0 )  
                a += ", " + as_string( params.cArgs-argnr ) + ". Parameter";
            throw_com( hr, "IDispatch::Invoke", a.c_str() );
        }


        switch( result.vt )
        {
            case VT_EMPTY:      break;  
            case VT_NULL:       break;  // NULL ist auch undefined?

            case VT_ERROR:      if( result.scode != DISP_E_PARAMNOTFOUND )    // Das liefert PerlScript, wenn Funktion keine Rckgabe hat.
                                    throw_com( result.scode, "IDispatch::Invoke (result)", method_name.c_str() );
                                break;

            case VT_UNKNOWN:    string_result = printf_string( "IUnknown(0x%X)", (long)V_UNKNOWN(&result) );
                                result_type = rt_string; 
                                break;

            case VT_DISPATCH:   
            {
                IDispatch* idispatch = V_DISPATCH(&result);

                string expression = "new " Z_PERL_IDISPATCH_PACKAGE_NAME "(" + as_string( (int)idispatch ) + ")";
                sv_result = eval_pv( expression.c_str(), TRUE );
                result_type = rt_sv;

                if( idispatch )
                {
                    //fprintf( stderr, "perl_com.cxx: %x->AddRef()\n", (void*)idispatch );
                    idispatch->AddRef();  // DESTROY ruft Release()
                }

                break;
            }

            case VT_INT:        int_result = V_INT ( &result ); result_type = rt_int; break;
          //case VT_UINT:       int_result = V_UINT( &result ); result_type = rt_int; break;
            case VT_I1:         int_result = V_I1  ( &result ); result_type = rt_int; break;
            case VT_UI1:        int_result = V_UI1 ( &result ); result_type = rt_int; break;
            case VT_I2:         int_result = V_I2  ( &result ); result_type = rt_int; break;
            case VT_UI2:        int_result = V_UI2 ( &result ); result_type = rt_int; break;
            case VT_I4:         int_result = V_I4  ( &result ); result_type = rt_int; break;
          //case VT_UI4:        int_result = V_UI4 ( &result ); result_type = rt_int; break;

            case VT_R4:         double_result = V_R4( &result );  result_type = rt_double; break;
            case VT_R8:         double_result = V_R8( &result );  result_type = rt_double; break;

            case VT_BSTR:       string_result = string_from_bstr( V_BSTR( &result ) );
                                result_type = rt_string; 
                                break;

            default:            string_result = string_from_variant( result );
                                result_type = rt_string; 
                                break;
        }
    }
    catch( const exception&  x ) { error = true;  error_text = x.what(); }
    catch( const _com_error& x ) { error = true;  error_text = string_from_bstr( x.Description() ); }

    if( error )  croak( error_text.c_str() );   // ACHTUNG: croak() kehrt nicht zurck!
    //if( error )  result_type = rt_string, string_result = ERROR_ID + error_text;

    //dXSTARG;

    switch( result_type )
    {
        case rt_void: 
            XSRETURN_UNDEF;

        case rt_double:
            //XSprePUSH; 
            //PUSHn( double_result );
            //XSRETURN( 1 );
            XSRETURN_NV( double_result );

        case rt_int:
            //XSprePUSH; 
            //PUSHi( int_result );
            //XSRETURN( 1 );
            XSRETURN_IV( int_result );

        case rt_string:
            //XSprePUSH; 
            //PUSHp( string_result.c_str(), string_result.length() );
            //XSRETURN( 1 );
            XSRETURN_PVN( string_result.c_str(), string_result.length() );

        case rt_sv:
            dXSTARG;
            XSprePUSH; 
            PUSHs( sv_result ); 
            XSRETURN( 1 );
    }
}

//------------------------------------------------------------------------------------perl_com_boot

extern "C" 
XS( perl_com_boot )
{
    dXSARGS;
    unused(items);

    XS_VERSION_BOOTCHECK;

    newXS( Z_PERL_IDISPATCH_PACKAGE_NAME "::__destroy"     , XS_Zschimmer_destroy     , __FILE__ );
    newXS( Z_PERL_IDISPATCH_PACKAGE_NAME "::__resolve_name", XS_Zschimmer_resolve_name, __FILE__ );
    newXS( Z_PERL_IDISPATCH_PACKAGE_NAME "::__call"        , XS_Zschimmer_call        , __FILE__ );

    XSRETURN_YES;
}

//-------------------------------------------------------------------------------------------------

} //namespace zschimmer

