// $Id: xml_libxml2.cxx 3995 2005-10-31 07:57:11Z jz $

// Dokumentation von libxml2 in http://xmlsoft.org/html/libxml-lib.html

#include "zschimmer.h"
#include "xml_libxml2.h"
#include "mutex.h"
#include "z_com.h"
#include "threads.h"

#include <stdarg.h>

#include "../3rd_party/libxml2/libxml.h"
#include "../3rd_party/libxml2/include/libxml/xmlmemory.h"
#include "../3rd_party/libxml2/include/libxml/xmlIO.h"
#include "../3rd_party/libxml2/include/libxml/xpath.h"


using namespace std;
using namespace zschimmer::com;


namespace zschimmer {
namespace xml_libxml2 {
namespace xml {

//-------------------------------------------------------------------------------------------static

static Mutex                    libxml_parser_mutex ( "libxml_parser_mutex" );
static string                   static_parser_messages;
//static Thread_data<Libxml2_thread_data> thread_data;

//--------------------------------------------------------------------------------------------const

static Error_code_text error_codes[] =
{
    { "LIBXML2-001", "Fehler bei $1" },
    { "LIBXML2-002", "Knoten $1 erwartet, an der Stelle ist aber $2: $3" },
    { "LIBXML2-003", "Attribut $1 fehlt" },
    { "LIBXML2-004", "XML-Dokument ist nicht valide: $1" },
    { "LIBXML2-005", "XML-Element <$1> fehlt" },
    { NULL         , NULL }
};

//-------------------------------------------------------------------------------------------Z_INIT

Z_INIT( xml_libxml2 )
{
    add_error_code_texts( error_codes );
}

//----------------------------------------------------------------------------string_from_DOMString
/*
inline string string_from_DOMString( const xmlChar* s )
{
    return s? (const char*)s : "";
}
*/
//-------------------------------------------------------------------------------------------sd_free

string sd_free( xmlChar* s )
{
    string result = sd( s );
    if( s )  xmlFree( s );
    return result;
}

//----------------------------------------------------------------------------DOMString_from_string
/*
inline const xmlChar* DOMString_from_string( const string& s )
{
    return (const xmlChar*)s.c_str();
}
*/
//-----------------------------------------------------------------------------------------------ds
/*
inline const xmlChar* ds( const string& s )
{
    //return (const xmlChar*)s.c_str();
}
*/
//-----------------------------------------------------------------------------------------------ds
/*
inline const xmlChar* ds( const char* s )
{
    return (const xmlChar*)( s? s : "" );
}
*/
//----------------------------------------------------------------------------------xml_print_error

static void xml_print_error( void* context, const char* format, ... )
{
    char buf[1024];
    va_list args;

    va_start( args, format );
    int ret = _vsnprintf( buf, sizeof(buf), format, args ); 
    va_end( args );

    if( ret > 0 )  ((Libxml2_error_text*)context)->_error_text.append( buf, ret );
    //if( ret > 0 )  thread_data->_error_text.append( buf, ret );
}

//---------------------------------------------------------Activate_error_text::Activate_error_text

Activate_error_text::Activate_error_text( Libxml2_error_text* object )
{
    object->_error_text = "";
    xmlSetGenericErrorFunc( object, (xmlGenericErrorFunc)xml_print_error );
}

//--------------------------------------------------------Activate_error_text::~Activate_error_text

Activate_error_text::~Activate_error_text()
{
    xmlSetGenericErrorFunc( NULL, NULL );
}
//-----------------------------------------------------------------------Utf8_string::set_iso8859_1

void Utf8_string::set_iso8859_1( const char* str, int length_hint )
{
    _utf8 = "";
    _utf8.reserve( length_hint + length_hint / 10 );        // 10% Umlaute angenommen

    const char* s = str;

    while(1)
    {
        const char* s0 = s;
        while( *s  &&  (unsigned char)*s < 0x80 )  s++;
        _utf8.append( s0, s - s0 );
    
        if( *s == '\0' )  return;

        while( (unsigned char)*s >= 0x80 )
        {
            _utf8 += 0xC0 + ( (unsigned char)*s >> 6   );
            _utf8 += 0x80 + ( (unsigned char)*s & 0x3F );
            if( *++s == '\0' )  return;
        }
    }
}

//------------------------------------------------------------------------------------Print_context

struct Print_context
{
    static void static_print( void* context, const char *format_msg, ... )
    {
        va_list list;
        va_start( list, format_msg );

        ((Print_context*)context)->print( format_msg, list );

        va_end( list );
    }

    void print( const char* format_msg, va_list list )
    {
        vfprintf( stderr, format_msg, list );

        char buffer [1024];
        _vsnprintf( buffer, sizeof buffer, format_msg, list );
        _messages += buffer;
    }

    string  _messages;
};

//-----------------------------------------------------------------------------------------------sd

string sd( const xmlChar* utf8 )
{
    string result;

    if( utf8 )
    {
        int char_count = xmlUTF8Strlen( utf8 );
        // Optimierung, wenn Lnge von utf8 bekannt: if( char_count == utf8.length() )  return string( utf8 );  // Nur ASCII-Zeichen
        result.reserve( char_count );

        const xmlChar* u = utf8;

        while( *u )
        {
            const xmlChar* u0 = u;
            while( *u != 0  &&  *u < 0x80 )  u++;
            result.append( (const char*)u0, u - u0 );

            while( *u >= 0x80 )
            {
                uint c = 0;

                if( *u < 0x80 )  c += *u++;
                else
                if( *u < 0xE0 )  c += ( ( u[0] & ~0xC0 ) <<  6 ) |   ( u[1] & 0x3F )                                                     ,  u += 2;
                else
                if( *u < 0xF0 )  c += ( ( u[0] & ~0xE0 ) << 12 ) | ( ( u[1] & 0x3F ) <<  6 ) |   ( u[2] & 0x3F )                         ,  u += 3;
                else
                                 c += ( ( u[0] & ~0xF0 ) << 18 ) | ( ( u[1] & 0x3F ) << 12 ) | ( ( u[2] & 0x3F ) << 6 ) | ( u[3] & 0x3F ),  u += 4;

                result += c <= 0xFF? (char)c : '';
            }
        }
    }

    return result;
}

//--------------------------------------------------------------------------------name_of_node_type

string name_of_node_type( const NodeType& n )
{
    switch( n )
    {
        case ELEMENT_NODE                    : return "ELEMENT_NODE";                   break;
        case ATTRIBUTE_NODE                  : return "ATTRIBUTE_NODE";                 break;
        case TEXT_NODE                       : return "TEXT_NODE";                      break;
        case CDATA_SECTION_NODE              : return "CDATA_SECTION_NODE";             break;
        case ENTITY_REFERENCE_NODE           : return "ENTITY_REFERENCE_NODE";          break;
        case ENTITY_NODE                     : return "ENTITY_NODE";                    break;
        case PROCESSING_INSTRUCTION_NODE     : return "PROCESSING_INSTRUCTION_NODE";    break;
        case COMMENT_NODE                    : return "COMMENT_NODE";                   break;
        case DOCUMENT_NODE                   : return "DOCUMENT_NODE";                  break;
        case DOCUMENT_TYPE_NODE              : return "DOCUMENT_TYPE_NODE";             break;
        case DOCUMENT_FRAGMENT_NODE          : return "DOCUMENT_FRAGMENT_NODE";         break;
        case NOTATION_NODE                   : return "NOTATION_NODE";                  break;
                                     default : return as_string( (int)n ); 
    }
};

//----------------------------------------------------------------------------------------sax_print
// Wird mit gesperrtem Mutex libxml_parser_mutex gerufen.
// Schnell raus hier!
/*
static void sax_print( void*, const char *format_msg, ... )
{
    va_list list;

    va_start( list, format_msg );

    vfprintf( stderr, format_msg, list );

    char buffer [1024];
    int ret = _vsnprintf( buffer, sizeof buffer, format_msg, list );
    static_parser_messages += buffer;

    va_end( list );
}
*/
//---------------------------------------------------------------------------------Dtd_ptr::release

void Dtd_ptr::release()
{ 
    if( _ptr )
    {
        xmlFreeDtd( _ptr );
        _ptr = NULL; 
    }
}

//------------------------------------------------------------------------------------Dtd_ptr::read

void Dtd_ptr::read( const string& dtd_string )
{
    release();

    /*
    XML_CHAR_ENCODING_8859_1, XML_CHAR_ENCODING_UTF8 
    xmlIOParseDTD		(xmlSAXHandlerPtr sax, 					 xmlParserInputBufferPtr input, 					 xmlCharEncoding enc)

    xmlParserCtxtPtr xmlCreateIOParserCtxt	(xmlSAXHandlerPtr sax,
					 void *user_data,
					 xmlInputReadCallback   ioread,
					 xmlInputCloseCallback  ioclose,
					 void *ioctx,
					 xmlCharEncoding enc);

    */

    xmlParserInputBufferPtr input_buffer = xmlParserInputBufferCreateMem( dtd_string.data(), dtd_string.length(), XML_CHAR_ENCODING_NONE ); 
    if( !input_buffer )  throw_xc( "LIBXML2-001", "xmlParserInputBufferCreateMem()" );


    /*
    xmlSAXHandler sax_handler;
    memset( &sax_handler, 0, sizeof sax_handler );

    sax_handler.fatalError  = (fatalErrorSAXFunc)sax_print;
    sax_handler.error       = (errorSAXFunc)     sax_print;
    sax_handler.warning     = (warningSAXFunc)   sax_print;
    sax_handler.initialized = 1;
    */

    //Z_MUTEX( libxml_parser_mutex )
    {
        //static_parser_messages = "";

        _ptr = xmlIOParseDTD( NULL, input_buffer, XML_CHAR_ENCODING_NONE ); 

        if( !_ptr )  throw_xc( "LIBXML2-001", "xmlIOParseDTD()", static_parser_messages );
    }
    //Offenbar von xmlIOParseDTD() erledigt: xmlFreeParserInputBuffer( input_buffer );

}

//-----------------------------------------------------------------------Document_ptr::Document_ptr
/* nicht geprft
Document_ptr::Document_ptr( const Element_ptr& root_to_clone )
{
    create();
    appendChild( clone( root_to_clone ) );
}
*/
//----------------------------------------------------------------------------Document_ptr::release

void Document_ptr::release()
{ 
    if( _ptr ) 
    {
        if( --ptr()->zschimmer_ref_count == 0 )  
        {
            xmlFreeDoc( ptr() );
            _ptr = NULL; 
        }
    }
}

//-----------------------------------------------------------------------------Document_ptr::assign

void Document_ptr::assign( xmlDoc* doc )
{ 
    release(); 
    _ptr = (xmlNode*)doc;  
    if( doc  )  doc->zschimmer_ref_count++; 
}

//-----------------------------------------------------------------------------Document_ptr::create

void Document_ptr::create()
{ 
    release();

    _ptr = (xmlNode*)xmlNewDoc( (const xmlChar*)"1.0" );  
    if(!_ptr)  throw_xc( "LIBXML2-001", "xmlNewDoc()" );  
    ptr()->zschimmer_ref_count = 1;        // Erweiterung von struct xmlDoc von mir, Joacim Zschimmer.
}

//-----------------------------------------------------------------------Document_ptr::try_load_xml

bool Document_ptr::try_load_xml( const string& xml_text )
{
    xmlDocPtr doc = NULL;

    Z_MUTEX( libxml_parser_mutex )
    {
        release();
    
        xmlKeepBlanksDefault(0);     // Das ist ja eine statische Funktion!
        xmlLineNumbersDefault( 1 );        // Zeilennummern zhlen

        Activate_error_text activate_error_text ( this );

        //xmlParserCtxtPtr parser_context = xmlCreateDocParserCtxt( (xmlChar*)xml_text.c_str() );
        //xmlFreeParserCtxt( parser_context );
        doc = xmlReadMemory( xml_text.data(), xml_text.length(), "", NULL, 
                             XML_PARSE_DTDLOAD |    // load the external subset
                             XML_PARSE_DTDATTR |    // default DTD attributes
                             XML_PARSE_DTDVALID     // validate with the DTD
                         //? XML_PARSE_NOBLANKS |   // remove blank nodes
                         //? XML_PARSE_XINCLUDE |   // Implement XInclude substitition
                           );

        //doc = xmlParseDoc( (xmlChar*)xml_text.c_str() );
        _ptr = (xmlNode*)doc;

        xmlSetGenericErrorFunc( NULL, NULL );
        xmlCleanupParser();
    }


    if( !_ptr )  return false;



    ptr()->zschimmer_ref_count = 1;        // Erweiterung von struct xmlDoc von mir, Joacim Zschimmer.


    // VALIDIERUNG

    if( doc->intSubset || doc->extSubset )
    {
        Print_context print_context;
        xmlValidCtxt validation_context;

        validation_context.userData = (void*)                  &print_context;
        validation_context.error    = (xmlValidityErrorFunc)   Print_context::static_print;
        validation_context.warning  = (xmlValidityWarningFunc) Print_context::static_print;

        int ok = xmlValidateDocument( &validation_context, doc );
        if( !ok )  throw_xc( "LIBXML2-004", print_context._messages );
    }

    return true;
}

//-----------------------------------------------------------------------Document_ptr::try_load_xml

bool Document_ptr::try_load_xml( const BSTR xml_text_bstr )
{
    return try_load_xml( string_from_bstr( xml_text_bstr ) );
}

//---------------------------------------------------------------------------Document_ptr::load_xml

void Document_ptr::load_xml( const string& xml_text )
{
    bool ok = try_load_xml( xml_text );
    if( !ok )  throw_xc( "LIBXML2", error_text() );
}

//---------------------------------------------------------------------------Document_ptr::load_xml

void Document_ptr::load_xml( const BSTR xml_text )
{
    bool ok = try_load_xml( xml_text );
    if( !ok )  throw_xc( "LIBXML2", error_text() );
}

//--------------------------------------------------------Document_ptr::validate_against_dtd_string

void Document_ptr::validate_against_dtd_string( const string& dtd_string )
{
    Dtd_ptr dtd ( dtd_string );
    validate_against_dtd( dtd );
}

//---------------------------------------------------------------Document_ptr::validate_against_dtd

void Document_ptr::validate_against_dtd( xmlDtdPtr dtd )
{
    Print_context print_context;
    xmlValidCtxt  validation_context;

    validation_context.userData = (void*)                  &print_context;
    validation_context.error    = (xmlValidityErrorFunc)   Print_context::static_print;
    validation_context.warning  = (xmlValidityWarningFunc) Print_context::static_print;

    int ok = xmlValidateDtd( &validation_context, (xmlDocPtr)_ptr, dtd );
    if( !ok )  throw_xc( "LIBXML2-004", print_context._messages );
}

//--------------------------------------------------------------------------------Document_ptr::xml

string Document_ptr::xml( bool indented ) const
{ 
    string   result;
    xmlChar* buffer = NULL;
    int      length = 0;

    if( indented )  xmlDocDumpFormatMemory( ptr(), &buffer, &length, 1 );
              else  xmlDocDumpMemory      ( ptr(), &buffer, &length );


    // Doppeltes <?xml?> entfernen
    
    const char* p = (const char*)buffer;

    if( length > 1+5 )
    {
        const char* p2 = (const char*)memchr( p+1, '<', length - 5 );
        if( p2 &&  memcmp( p2, "<?xml", 5 ) == 0 )  p = p2;
    }


    result = make_string( p, (const char*)buffer + length - p );

    xmlFree( buffer );

    return result;
}

//----------------------------------------------------------------------Document_ptr::createElement

Element_ptr Document_ptr::createElement( const string& tagName ) const
{ 
    return (Element_ptr)createElement( tagName.c_str() ); 
}

//----------------------------------------------------------------------Document_ptr::createElement

Element_ptr Document_ptr::createElement( const char* tagName ) const
{ 
    return (Element_ptr) xmlNewDocNode( ptr(), NULL, Utf8_string(tagName).utf8(), NULL ); 
}

//---------------------------------------------------------------------Document_ptr::createTextNode

Text_ptr Document_ptr::createTextNode( const string& data ) const
{ 
    return xmlNewDocText( ptr(), Utf8_string(data).utf8() ); 
}

//----------------------------------------------------------------------Document_ptr::createComment

Comment_ptr Document_ptr::createComment( const string& data ) const            
{ 
    return createComment( data.c_str() ); 
}

//----------------------------------------------------------------------Document_ptr::createComment

Comment_ptr Document_ptr::createComment( const char*   data ) const
{ 
    return xmlNewComment( Utf8_string(data).utf8() ); 
}

//-----------------------------------------------------------------Document_ptr::createCDATASection

CDATASection_ptr Document_ptr::createCDATASection( const string& data ) const            
{ 
    Utf8_string u = data; 
    return xmlNewCDataBlock( ptr(), u.utf8(), u.byte_count() ); 
}

//--------------------------------------------------------Document_ptr::createProcessingInstruction

ProcessingInstruction_ptr Document_ptr::createProcessingInstruction( const string& target, const string& data ) const      
{ 
    return xmlNewPI( Utf8_string(target).utf8(), Utf8_string(data).utf8() ); 
}

//--------------------------------------------------------------------Document_ptr::documentElement

Element_ptr Document_ptr::documentElement() const                                
{ 
    return (Element_ptr)xmlDocGetRootElement( ptr() ); 
}

//----------------------------------------------------------------Document_ptr::create_root_element

Element_ptr Document_ptr::create_root_element( const string& name )
{ 
    return appendChild( createElement( name ) ); 
}

//------------------------------------------------------------------------------Document_ptr::clone

Node_ptr Document_ptr::clone( const Node_ptr& node, int extended ) const
{ 
    xmlNodePtr result = xmlDocCopyNode( node.ptr(), ptr(), extended );

    if( !result )  throw_xc( "LIBXML2-001", "xmlDocCopyNode" );

    return result;
}

//---------------------------------------------------------------------------------Node_ptr::assign

void Node_ptr::assign( _xmlNode* ptr, NodeType type )
{ 
    _ptr = ptr; 

    if( _ptr )
    {
        try
        {
            assert_type( type ); 
        }
        catch( const exception& )
        {
            _ptr = NULL;
            throw;
        }
    }
}

//----------------------------------------------------------------------------Node_ptr::assert_type

void Node_ptr::assert_type( NodeType type )
{ 
    if( !is_type( type ) )  throw_node_type( type ); 
}

//---------------------------------------------------------------------------Node_ptr::set_nodeName

void Node_ptr::set_nodeName( const string& name ) const
{ 
    xmlNodeSetName( _ptr, (xmlChar*)name.c_str() ); 
}

//-------------------------------------------------------------------------------Node_ptr::nodeName

string Node_ptr::nodeName() const
{ 
    return sd( _ptr->name ); 
}

//----------------------------------------------------------------------------Node_ptr::nodeName_is

bool Node_ptr::nodeName_is( const char* name ) const
{ 
    return strcmp( (const char*)_ptr->name, name ) == 0; 
}

//------------------------------------------------------------------------------Node_ptr::nodeValue

string Node_ptr::nodeValue() const
{ 
    return sd_free( xmlNodeGetContent( _ptr ) ); 
}

//-------------------------------------------------------------------------------Node_ptr::nodeType

NodeType Node_ptr::nodeType() const
{ 
    return (NodeType)_ptr->type; 
}

//-----------------------------------------------------------------------------Node_ptr::parentNode

Node_ptr Node_ptr::parentNode() const
{ 
    return _ptr->parent; 
}

//-----------------------------------------------------------------------------Node_ptr::childNodes
/*
NodeList_ptr Node_ptr::childNodes() const
{
}
*/
//-----------------------------------------------------------------------------Node_ptr::firstChild

Node_ptr Node_ptr::firstChild() const
{ 
    return _ptr->children; 
}

//------------------------------------------------------------------------------Node_ptr::lastChild

Node_ptr Node_ptr::lastChild() const
{ 
    return xmlGetLastChild( _ptr ); 
}

//------------------------------------------------------------------------Node_ptr::previousSibling

Node_ptr Node_ptr::previousSibling() const
{ 
    return _ptr->prev; 
}

//----------------------------------------------------------------------------Node_ptr::nextSibling

Node_ptr Node_ptr::nextSibling() const
{ 
    return _ptr->next; 
}

//---------------------------------------------------------------------------Node_ptr::replaceChild

Node_ptr Node_ptr::replaceChild( Node_ptr newChild, Node_ptr oldChild ) const 
{ 
    return xmlReplaceNode( oldChild, newChild ); 
}

//----------------------------------------------------------------------------Node_ptr::appendChild

Node_ptr Node_ptr::appendChild( Node_ptr newChild ) const
{ 
    return xmlAddChild( _ptr, newChild ); 
}

//--------------------------------------------------------------------------Node_ptr::hasChildNodes

bool Node_ptr::hasChildNodes() const
{ 
    return _ptr->children != NULL; 
}

//-------------------------------------------------------------------------Node_ptr::getTextContent

string Node_ptr::getTextContent() const
{ 
    return sd_free( xmlNodeGetContent( _ptr ) ); 
}

//----------------------------------------------------------------------------Node_ptr::line_number

int Node_ptr::line_number() const
{ 
    return XML_GET_LINE( _ptr ); 
}

//------------------------------------------------------------------------Node_ptr::throw_node_type    

void Node_ptr::throw_node_type( NodeType type ) const
{
    Xc x ( "LIBXML2-002" );

    x.insert( 1, name_of_node_type( type ) );
    x.insert( 2, name_of_node_type( nodeType() ) );
    x.insert( 3, nodeName() );

    x.append_text( getTextContent().substr( 0, 100 ) );

    throw_xc( x );
}

//-----------------------------------------------------------------------------Node_ptr::childNodes

NodeList_ptr Node_ptr::childNodes() const
{ 
    return _ptr->children; 
}

//--------------------------------------------------------------------------Node_ptr::ownerDocument

Document_ptr Node_ptr::ownerDocument() const
{ 
    return _ptr->doc; 
}

//------------------------------------------------------------------------------Node_ptr::cloneNode

Node_ptr Node_ptr::cloneNode( bool deep ) const
{ 
    xmlNodePtr result = xmlCopyNode( _ptr, deep? 1 : 2 ); 
    if( !result )  throw_xc( "xmlCopyNode" );

    return result;
}

//---------------------------------------------------------------------------Node_ptr::select_nodes

Xpath_nodes Node_ptr::select_nodes( const string& xpath_expression ) const
{
    Xpath_nodes result;
    
    if( _ptr )
    {
        result._xpath_object = Z_NEW( Xpath_nodes::Xpath_object );

        result._xpath_object->_xpath_context = xmlXPathNewContext( _ptr->doc );
        if( !result._xpath_object->_xpath_context )  throw_xc( "LIBXML2-001", "xmlXPathNewContext" );

        result._xpath_object->_xpath_context->node = _ptr;

        result._xpath_object->_xpath_object = xmlXPathEvalExpression( Utf8_string( xpath_expression ).utf8(), result._xpath_object->_xpath_context );
        if( !result._xpath_object->_xpath_object )  throw_xc( "LIBXML2-001", "xmlXPathEvalExpression", xpath_expression );
    }

    return result;
}

//----------------------------------------------------------------------------Node_ptr::select_node

Node_ptr Node_ptr::select_node( const string& xpath_expression ) const
{
    Xpath_nodes nodes = select_nodes( xpath_expression );
    return nodes.count() > 0? nodes[0] : NULL;
}

//-------------------------------------------------------------------------------Node_ptr::has_node

bool Node_ptr::has_node( const string& xpath_expression ) const
{
    return select_node( xpath_expression ) != NULL;
}

//---------------------------------------------------------------------Node_ptr::select_node_strict

Node_ptr Node_ptr::select_node_strict( const string& xpath_expression ) const
{
    Node_ptr result = select_node( xpath_expression );
    if( !result )  throw_xc( "LIBXML2-005", xpath_expression );
    return result;
}

//---------------------------------------------------------Xpath_nodes::Xpath_object::~Xpath_object

Xpath_nodes::Xpath_object::~Xpath_object()
{
    xmlXPathFreeObject( _xpath_object );
    xmlXPathFreeContext( _xpath_context ); 
}

//-------------------------------------------------------------------------------Xpath_nodes::count

int Xpath_nodes::count() const
{
    return _xpath_object &&  _xpath_object->_xpath_object->nodesetval? _xpath_object->_xpath_object->nodesetval->nodeNr : 0;
}

//---------------------------------------------------------------------------------Xpath_nodes::get

Node_ptr Xpath_nodes::get( int i ) const
{
    xmlNodePtr result = _xpath_object->_xpath_object->nodesetval->nodeTab[ i ];

    assert( result->type == ELEMENT_NODE );

    return result;
}

//------------------------------------------------------------------------Element_ptr::setAttribute

void Element_ptr::setAttribute( const string& name, const char* value ) const   
{ 
    xmlNewProp( _ptr, Utf8_string(name).utf8(), Utf8_string(value).utf8() ); 
}

//------------------------------------------------------------------------Element_ptr::getAttribute

string Element_ptr::getAttribute( const string& name, const string& deflt ) const  
{ 
    return getAttribute( name.c_str(), deflt.c_str() ); 
}

//--------------------------------------------------------------------Element_ptr::getAttributeNode

Attr_ptr Element_ptr::getAttributeNode( const char* name ) const
{ 
    Attr_ptr attr = xmlHasProp( _ptr, Utf8_string(name).utf8() ); 

    // ACHTUNG:
    // Bei DOCTYPE: Fehlendes Attribut wird als "" geliefert, wenn es als #IMPLIED deklariert ist.
    // ATTRIBUT="" WIRKT WIE EIN NICHT ANGEGEBENES ATTRIBUT!

    if( attr  &&  attr.value() == "" )  attr = NULL;        

    return attr;
}

//------------------------------------------------------------------------Element_ptr::getAttribute

string Element_ptr::getAttribute( const char* name, const char* deflt ) const  
{ 
    Attr_ptr a = getAttributeNode( name ); 
    return a? a.value() : deflt; 

/*
    xmlNode* cur = _ptr->xmlChildrenNode;

	while (cur != NULL) {
	    if ((!xmlStrcmp(cur->name, (const xmlChar *)"reference"))) {
	 printf("uri: %s\n", xmlGetProp(cur, "uri"));
		    }
	    cur = cur->next;
	}
*/

}

//-------------------------------------------------------------------Element_ptr::bool_getAttribute

bool Element_ptr::bool_getAttribute( const string& name, bool deflt ) const
{
    Attr_ptr attr = getAttributeNode( name );
    if( attr )  return as_bool( attr.value() );
          else  return deflt;
}

//--------------------------------------------------------------------Element_ptr::int_getAttribute

int Element_ptr::int_getAttribute( const string& name ) const
{
    Attr_ptr attr = getAttributeNode( name );
    if( attr )  return as_int( attr.value() );
          else  throw_xc( "LIBXML2-003", name );
}

//--------------------------------------------------------------------Element_ptr::int_getAttribute

int Element_ptr::int_getAttribute( const string& name, int deflt ) const
{
    Attr_ptr attr = getAttributeNode( name );
    if( attr )  return as_int( attr.value() );
          else  return deflt;
}

//--------------------------------------------------------------------Element_ptr::int_getAttribute

int Element_ptr::uint_getAttribute( const string& name, int deflt ) const
{
    Attr_ptr attr = getAttributeNode( name );
    if( attr )  return as_uint( attr.value() );
          else  return deflt;
}

//------------------------------------------------------------------------Element_ptr::setAttribute

void Element_ptr::setAttribute( const string& name, bool value ) const
{
    setAttribute( name, as_string( value ) );
}

//------------------------------------------------------------------------Element_ptr::setAttribute

void Element_ptr::setAttribute( const string& name, int64 value ) const
{
    setAttribute( name, as_string( value ) );
}

//------------------------------------------------------------------------Element_ptr::hasAttribute

bool Element_ptr::hasAttribute( const string& name ) const
{
    return xmlHasProp( _ptr, Utf8_string(name).utf8() ) != NULL;
}

//------------------------------------------------------------------Element_ptr::append_new_element

Element_ptr Element_ptr::append_new_element( const string& element_name )
{
    xml::Element_ptr element = ownerDocument().createElement( element_name );
    appendChild( element );
    return element;
}

//-------------------------------------------------------------Element_ptr::append_new_text_element

Element_ptr Element_ptr::append_new_text_element( const string& element_name, const string& text )
{
    xml::Element_ptr element = append_new_element( element_name );
    element.appendChild( ownerDocument().createTextNode( text ) );
    return element;
}

//----------------------------------------------Element_ptr::append_new_name_value_element_optional

Element_ptr Element_ptr::append_new_name_value_element_optional( const string& element_name, 
                                                                 const string& name_attribute_name, const string& name_value, 
                                                                 const string& value_attribute_name, const string& value_value )
{
    return append_new_name_value_element_optional( element_name.c_str(), 
                                                   name_attribute_name.c_str(), name_value.c_str(), 
                                                   value_attribute_name.c_str(), value_value );
}

//----------------------------------------------Element_ptr::append_new_name_value_element_optional

Element_ptr Element_ptr::append_new_name_value_element_optional( const char* element_name, 
                                                                 const char* name_attribute_name, const char* name_value, 
                                                                 const char* value_attribute_name, const string& value_value )
{
    Element_ptr result;

    if( value_value != "" )
    {
        result = append_new_element( element_name );
        result.setAttribute( name_attribute_name , name_value  );
        result.setAttribute( value_attribute_name, value_value );
    }

    return result;
}

//------------------------------------------------------------------Element_ptr::append_new_comment

Comment_ptr Element_ptr::append_new_comment( const string& comment )
{
    Comment_ptr result = ownerDocument().createComment( comment );
    appendChild( result );
    return result;
}

//---------------------------------------------------------------------------------Element_ptr::xml
/*
string Element_ptr::xml( bool indented ) const
{ 
    string   result;
    xmlChar* buffer = NULL;
    int      length = 0;

    if( indented )  xmlNodeDump( ptr(), &buffer, &length, 1 );
              else  xmlElemDump( ptr(), &buffer, &length );


    // Doppeltes <?xml?> entfernen
    
    const char* p = (const char*)buffer;

    if( length > 1+5 )
    {
        const char* p2 = (const char*)memchr( p+1, '<', min( 400, length - 5 ) );
        if( p2 &&  memcmp( p2, "<?xml", 5 ) == 0 )  p = p2;
    }


    result = make_string( p, (const char*)buffer + length - p );

    xmlFree( buffer );

    return result;
}
*/
//-------------------------------------------------------------------------------Document_ptr::text

string Element_ptr::text() const
{
    if( !this  ||  !_ptr )  return "";


    list<string> result;

    for( xml::Node_ptr node = firstChild(); node; node = node.nextSibling() )
    {
        switch( node.nodeType() )
        {
            case TEXT_NODE: 
                result.push_back( Text_ptr(node).data() );
                break;

            case CDATA_SECTION_NODE: 
                result.push_back( Text_ptr(node).data() );
                break;

            case ELEMENT_NODE:
                result.push_back( Element_ptr(node).text() );

            default: ;
        }
    }

    return join( "", result );
}

//-------------------------------------------------------------------------------Document_ptr::text

string Element_ptr::trimmed_text() const
{
    string text = this->text();

    const char* p0 = text.c_str();
    const char* p  = p0;
    while( *p == ' ' || *p == '\t' )  p++;            // Leerzeichen hinter <tag> lschen
    if( *p == '\r' )  p++;              // Ersten Zeilenwechsel lschen
    if( *p == '\n' )  p++;

    const char* p_end = p0 + text.length();
    while( p_end > p0  &&  ( p_end[-1] == ' ' || p_end[-1] == '\t' ) )  p_end--;    // Leerzeichen vor </tag> lschen
    if( p_end > p0  &&  p_end[-1] == '\n' )  p_end--;                           // Letzten Zeilenwechsel lschen
    if( p_end > p0  &&  p_end[-1] == '\r' )  p_end--;

    //while( p_end > p  &&  isspace( (uint)p_end[-1] ) )  p_end--;        // Alle Leerzeichen und Blanks etc. am Ende lshcen

    text.erase( p_end - p0 );
    text.erase( 0, p - p0 );

    return text;
}

//-----------------------------------------------------------------------------------Text_ptr::data

string Text_ptr::data() const
{ 
    return sd_free( xmlNodeGetContent( _ptr ) ); 
}

//----------------------------------------------------------------------------------Attr_ptr::value

string Attr_ptr::value()
{ 
    return sd_free( xmlNodeGetContent( _ptr->children ) ); 
}

//-------------------------------------------------------------------------------------------------

} //namespace xml
} //namespace xml_libxml2
} //namespace zschimmer

//-------------------------------------------------------------------------------------------------
