// $Id: z_windows.cxx 3668 2005-05-17 10:47:25Z jz $

#include "zschimmer.h"

#ifdef Z_WINDOWS

#include "log.h"

namespace zschimmer {
namespace windows {


//-------------------------------------------------------------------------------------------------

static Error_code_text error_codes[] =
{
    { "Z-SHOWWINDOW-01", "Ungltiger Wert fr ShowWindow(): " },
    {}
};

//-------------------------------------------------------------------------------------------Z_INIT

Z_INIT( z_windows )
{
    add_error_code_texts( error_codes ); 
}

//-----------------------------------------------------------------------------windows_message_step

void windows_message_step()
{
    MSG msg;

    while(1)
    {
        msg.message = WM_NULL;

        int ret = PeekMessage( &msg, NULL, 0, 0, PM_REMOVE );
        if( !ret )  break;
        if( msg.message == WM_NULL )  break;

        //TranslateMessage( &msg ); 
        DispatchMessage( &msg ); 
    }
}

//--------------------------------------------------------------------------------show_window_value

int show_window_value( const string& show_window_value_name )
{
    string s = ucase( show_window_value_name );
    if( string_begins_with( s, "SW_" ) )  s.erase( 0, 3 );

    if( s == "FORCEMINIMIZE"    )  return SW_FORCEMINIMIZE;     // Windows 2000/XP: Minimizes a window, even if the thread that owns the window is hung. This flag should only be used when minimizing windows from a different thread.
    if( s == "HIDE"             )  return SW_HIDE;              // Hides the window and activates another window.
    if( s == "MAXIMIZE"         )  return SW_MAXIMIZE;          // Maximizes the specified window.
    if( s == "MINIMIZE"         )  return SW_MINIMIZE;          // Minimizes the specified window and activates the next top-level window in the Z order.
    if( s == "RESTORE"          )  return SW_RESTORE;           // Activates and displays the window. If the window is minimized or maximized, the system restores it to its original size and position. An application should specify this flag when restoring a minimized window.
    if( s == "SHOW"             )  return SW_SHOW;              // Activates the window and displays it in its current size and position. 
    if( s == "SHOWDEFAULT"      )  return SW_SHOWDEFAULT;       // Sets the show state based on the SW_ value specified in the STARTUPINFO structure passed to the CreateProcess function by the program that started the application. 
    if( s == "SHOWMAXIMIZED"    )  return SW_SHOWMAXIMIZED;     // Activates the window and displays it as a maximized window.
    if( s == "SHOWMINIMIZED"    )  return SW_SHOWMINIMIZED;     // Activates the window and displays it as a minimized window.
    if( s == "SHOWMINNOACTIVE"  )  return SW_SHOWMINNOACTIVE;   // Displays the window as a minimized window. This value is similar to SW_SHOWMINIMIZED, except the window is not activated.
    if( s == "SHOWNA"           )  return SW_SHOWNA;            // Displays the window in its current size and position. This value is similar to SW_SHOW, except the window is not activated.
    if( s == "SHOWNOACTIVATE"   )  return SW_SHOWNOACTIVATE;    // Displays a window in its most recent size and position. This value is similar to SW_SHOWNORMAL, except the window is not actived.
    if( s == "SHOWNORMAL"       )  return SW_SHOWNORMAL;        // Activates and displays a window. If the window is minimized or maximized, the system restores it to its original size and position. An application should specify this flag when displaying the window for the first time.

    throw_xc( "Z-SHOWWINDOW-1", show_window_value_name );
}

//-------------------------------------------------------------------------------------------------

} //namespace windows
} //namespace zschimmer

#endif

